/*
Example custom stored proc resolver for Northwind Products table.
RHS:   9 Oct 1998

Publish the table Products for Merge replication, (if it is not already published).
Then go to the Publications Properties dialog in Enterprise Manager.  Go to the Articles tab, 
and expand the properties for the Products article (the button with "...").   
Choose the Resolver tab, and check the radio button for "Use This Custom Resolver",
selecting "Microsoft SQL Server stored procedure resolver" Then enter "sp_avgprice" 
(without the quotes) which is the name of the procedure below.

This sp calculates the average UnitPrice and works on the Products table in northwind 
1. It uses a distributed query to get the info from the Subscriber.
2. It computes the average price if the values are different and makes that the resolved value.
3. It logs the conflict at the publisher so it can be viewed and/or changed.
4. It finds the vice president of sales, and sends him email saying the price was
changed to an average due to a conflict.

IMPORTANT:
You must first set up a linked server for every subscriber that will merge so the 
distributed query can work.   Test that you can do a simple distributed query from the Publisher to 
Northwind.dbo.Products on the subscriber.  If you cannot do such a distributed query, then
you must fix the configuration before this example can work.
Here is an example of setting up the linked server and testing a distributed query

The Publisher is PUBSRV1.  The subscriber is SUBSRV:
*/

-- First add the subscriber as a linked server so can use distributed query.
-- Must also set up a linked srv login mapping.
-- Here is an example of how to do this - this runs at Publisher (PUBSRV1)
-- EXEC sp_addlinkedserver @server ='SUBSRV',@srvproduct ='SQL Server' 
-- EXEC sp_serveroption 'SUBSRV', 'data access', 'TRUE'

-- The following says that the login to the subscriber will be done via
-- the account of SQLUser and password of Password.  
-- exec sp_addlinkedsrvlogin 'SUBSRV','false',NULL,'SQLUser','Password'
-- Be sure 'SQLUser' or whatever mapping you use exists at the subscriber .

/*
-- Example.  This runs at SUBSRV.
exec sp_addlogin 'SQLUser','Password'
use Northwind
exec sp_adduser 'SQLUser'
GRANT SELECT ON Products to SQLUser
*/

-- Do a simple test of a distributed query from the Publisher PUBSRV1) to the Subscriber's 
-- (SUBSRV) Northwind.dbo.Products table
-- You must be able to make the following distributed query from your publisher to your 
-- subscriber, or the custom resolver won't work.  Do not continue until you have the basic 
-- plumbing working.

-- SELECT * FROM OPENQUERY([SUBSRV],'SELECT ProductID,UnitPrice,ROWGUIDCOL FROM Northwind.dbo.Products')

-- Ok, now that you have distributed query working, we can do the real work.
-- If you want to have the custom proc send mail, be sure 
-- sqlmail is started, or start it if not (uncomment the next line):
-- exec master..xp_startmail

use Northwind

IF EXISTS (select * from sysobjects where name='sp_avgprice')
	DROP PROCEDURE sp_avgprice
go

CREATE PROC sp_avgprice
@tableowner		sysname, -- this is the publisher owner, but we assume its always dbo at subscriber.
@tablename		sysname, -- this is the publisher tablename, but we assume its identical at subscriber.
@rowguid		uniqueidentifier,
@subscriber		sysname,
@subscriber_db		sysname,
@log_conflict		int OUTPUT, -- output param for if to log conflict for later resolution.  
@conflict_message 	nvarchar(512) OUTPUT -- output param for message to be given about resolution if conflict is loggged.

AS

DECLARE
@pub_qualified_name NVARCHAR(392), -- 3*128 + 2 spaces + 6 quotes 
@sub_qualified_name NVARCHAR(392),
@select_string NVARCHAR(2000),
@ProdID	int,
@SubUnitPrice money,
@rowguidvar uniqueidentifier,
@ResolvedUnitPrice money,
@PubUnitPrice money,
@mailsubject VARCHAR(80),
@recipients VARCHAR(80),
@ccline VARCHAR(80),
@bccline VARCHAR(80),
@mailtext VARCHAR(1000)

-- Unfortunately, you cannot parameterize SERVERNAME in a distributed query, including
-- OPENQUERY, so need to formulate a string to make the query to the subscriber.
-- You could put it into a temp table, but a FAST FORWARD cursor is better here.
-- First must fully qualify the table names - use QUOTENAME so can handle special chars.

-- This proc must exist at the PUBLISHER and in the database of the publication.
-- Note: The default stored proc resolver interface always expects the tablename to 
-- be the same between publisher and subscribers.   In this example, I assume the table
-- owner at the subscriber is 'dbo' regardless of who the publisher owner is.
-- Since this proc must reside in the publisher database, its really not essential 
-- to fully qualify with publisher db, but I will do so for illustration.

SELECT @pub_qualified_name=QUOTENAME(db_name())+'.'+QUOTENAME(@tableowner)+'.'+QUOTENAME(@tablename)
SELECT @sub_qualified_name=QUOTENAME(@subscriber_db)+'.'+QUOTENAME('dbo')+'.'+QUOTENAME(@tablename)

-- The permission on this stored procedure needs to be granted to public if users other 
-- than db_owner may run replication merge agents. Security check can be done so that only 
-- the users that are db_owner or in Publication Access List (PAL) can run it.
-- Uncomment the following and replace 'publication_name' with actual publication name 
-- if security check is needed. An error will be raised by sp_check_publication_access 
-- when security check fails. sp_check_publication_access should be run in the publishing
-- database.
/*
declare @retcode int
exec @retcode = sp_check_publication_access 
	@publication = 'publication_name'
if @retcode <> 0 or @@error <> 0
	RETURN(1)	
*/	

SELECT @select_string=
	'SELECT * FROM OPENQUERY('+QUOTENAME(@subscriber)+','+'"SELECT ProductID,UnitPrice,ROWGUIDCOL FROM ' + @sub_qualified_name +
	" WHERE ROWGUIDCOL='" + convert(varchar(36),@rowguid) + "'" + ' ") '

-- Note that we cannot use sp_executesql because rowguid is not an allowed datatype for a param
-- Rather than create a temp table to get values for variable assignment, 
-- will use a GLOBAL, FAST FORWARD CURSOR.

-- First make sure the cursor - since global - isn't still around from some failed prior use:
-- CURSOR_STATUS of -3 means the cursor does not exist, which is what we expect.
IF (CURSOR_STATUS('global','my_curs') <> -3)
	DEALLOCATE my_curs			-- old one still around - get rid of it.

SELECT @select_string='DECLARE my_curs CURSOR GLOBAL FAST_FORWARD FOR ' + @select_string
-- Go ahead and execute the declare cursor now, then we can fetch values into variables.
-- There will never be more than one row per guid - hence no FETCH loop necessary
EXECUTE (@select_string)
	if (@@error <> 0) 
		BEGIN	
		RAISERROR('Error Declaring Cursor.',18,-1)
		RETURN(1)
		END
OPEN my_curs
	if (@@error <> 0) 
		BEGIN
		RAISERROR('Error Opening Cursor.',18,-1)
		RETURN(1)
		END

FETCH NEXT FROM my_curs INTO @ProdID,@SubUnitPrice,@rowguidvar
	if (@@fetch_status <> 0)  -- We always expect first fetch to succeed.
		BEGIN
		RAISERROR('Error Fetching Subscriber row from Cursor.',18,-1)
		RETURN(1)
		END

FETCH NEXT FROM my_curs INTO @ProdID,@SubUnitPrice,@rowguidvar -- Since this FETCH will fail - variables not reassigned.
	if (@@fetch_status <> -1)  -- We always expect the second fetch to fail since we expect exactly one row.
		BEGIN
		RAISERROR('Error: Unexpected cursor result.  Multiple rows may have been retrieved from Subscriber.',18,-1)
		RETURN(1)
		END
CLOSE my_curs
DEALLOCATE my_curs

-- Ok, finally, we have the subscriber side values.   
-- Now a simple local query to the publisher table for its.  Since this 
-- proc exists only at the publisher, we will know the fully qualified name
-- so wont bother building up a string for it.

-- For the purposes of this example, we'll simply treat NULL as a 0 value , which may or may not
-- be desirable to do.

SELECT @PubUnitPrice=UnitPrice,
@ResolvedUnitPrice=( isnull(UnitPrice,0) + isnull(@SubUnitPrice,0) ) / 2.0 
FROM Products 
WHERE ROWGUIDCOL=@rowguid

-- If the prior select didn't successfully get exactly one row, there is a problem:
if (@@error <> 0 OR @@rowcount <> 1)
	BEGIN
	RAISERROR("Computation of Resolved Price Failed.",18,-1)	
	RETURN(1)
	END

-- If the resultant averaged price is no different than the subscribers value, then there
-- really was no conflict on UnitPrice.   Nothing needs to be done - we'd just keep the 
-- publisher values.   Otherwise, we will return the @ResolvedUnitPrice.  
-- Again, we'll treat NULL and zero as the same for this example.

IF ( isnull(@ResolvedUnitPrice,0) <> isnull(@SubUnitPrice,0) )
BEGIN

-- Return the resolved row - all publisher fields win except UnitPrice, which is the average.
SELECT 
ProductID,
ProductName,
SupplierID,
CategoryID,
QuantityPerUnit,
@ResolvedUnitPrice as UnitPrice,
UnitsInStock,
UnitsOnOrder,
ReorderLevel,
Discontinued,
rowguid                             
FROM Products
WHERE rowguid=@rowguid

-- Send mail to the VP of Sales.  Use TOP 1 in case more than 1, though if that really
-- were the case, you'd likely use a fast forward cursor to loop through and build the
-- @recipient list, with a ; as delimiter.
/*
SELECT @recipients=lower(FirstName)+'_'+lower(LastName)+'@hotmail.com'
	FROM (SELECT TOP 1 * FROM Employees WHERE lower(Title)='sales manager')as Emp
*/

-- Note that the prior SELECT will formulate an email name of steven_buchanan@hotmail.com if
-- the Northwind database has not been altered.   I've actually created that mail alias for this
-- example.   However, lets hardcode email alias to it in next line just in case the data was changed
-- so that we don't send email to some real account.
SELECT @recipients='steven_buchanan@hotmail.com'


SELECT @mailsubject='Price change conflict for Product ' + convert(varchar,@ProdID) + ' !'
SELECT @mailtext='The Price of product '+ convert(varchar,@ProdID) + ' was changed to $'
SELECT @mailtext=@mailtext + convert(varchar,@SubUnitPrice) + ' at computer ' + @subscriber
SELECT @mailtext=@mailtext + ', which is in conflict with the change to $' + convert(varchar,@PubUnitPrice)
SELECT @mailtext=@mailtext + ' made at computer ' + @@servername + '.   By rule, the average of the two, $'
SELECT @mailtext=@mailtext + convert(varchar,@ResolvedUnitPrice) + ', has been used.  You may '
SELECT @mailtext=@mailtext + 'view or change the conflict resolution using "View Conflicts" at '
SELECT @mailtext=@mailtext + @@servername + '.'

-- Also use that same message for the conflict message; this will show up in Conflict Viewer.
-- We elect here to log the publishing row as the conflict loser by virtue of @log_conflict=2;
-- set @log_conflict=1 to log subscribing row as the losing row.
SELECT @log_conflict=2,@conflict_message=@mailtext

SELECT @ccline=''  -- If you want people cc'd on the mail, add here.  Delimit with ;
SELECT @bccline='' -- If you want people bcc'd on the mail, addhere .  Delimit with ;

EXECUTE master..xp_sendmail @recipients=@recipients,
			@copy_recipients=@ccline,
			@blind_copy_recipients=@bccline,
			@subject=   @mailsubject,
                        @message=   @mailtext,                             	
                       	@no_output='true'

-- Not too much point to do error checking on the sendmail.  We've already resolved the conflict,
-- and so if the email send was NOT successful, we'd still just end which the proc is about to
-- do.
END

-- Note that if we did not meet the IF condition that UnitPrices differ , i.e. 
-- IF ( isnull(@ResolvedUnitPrice,0) <> isnull(@SubUnitPrice,0) )
-- we will have gotten here without doing really anything.   That 
-- means the default resolution will occur , i.e. the publisher's values win.

RETURN(0)
go
