if exists (select * from sysobjects where id = object_id('dbo.TActivity'))
	drop table dbo.TActivity
go
if exists (select * from sysobjects where id = object_id('dbo.TActivityTypes'))
	drop table dbo.TActivityTypes
go
if exists (select * from sysobjects where id = object_id('dbo.TAccounts'))
	drop table dbo.TAccounts
go
if exists (select * from sysobjects where id = object_id('dbo.TAccountTypes'))
	drop table dbo.TAccountTypes
go
if exists (select * from sysobjects where id = object_id('dbo.TCustomers'))
	drop table dbo.TCustomers
go

CREATE TABLE TCustomers
(
	CustomerID	int
		IDENTITY(1, 1)
		PRIMARY KEY CLUSTERED,
	CustomerName	varchar(30)	NOT NULL
		UNIQUE NONCLUSTERED,
	Optimist		timestamp	NOT NULL
)
go
GRANT INSERT, SELECT, UPDATE ON TCustomers TO PUBLIC
go

CREATE TABLE TAccountTypes
(
	AccountTypeID	int
		IDENTITY(1,1)
		PRIMARY KEY CLUSTERED,
	AccountType		varchar(15)	NOT NULL
		UNIQUE NONCLUSTERED,
	Optimist		timestamp	NOT NULL
)
go
GRANT SELECT, UPDATE ON TAccountTypes TO PUBLIC
go

CREATE TABLE TAccounts
(
	CustomerID		int		NOT NULL
		REFERENCES TCustomers(CustomerID),
	AccountTypeID	int		NOT NULL
		REFERENCES TAccountTypes(AccountTypeID)
		CONSTRAINT PK_TAccounts PRIMARY KEY CLUSTERED (CustomerID, AccountTypeID),
	OpenedOn		datetime 	NOT NULL
		DEFAULT(getdate()),
	Balance		money		NOT NULL
		DEFAULT(0),
	Optimist		timestamp	NOT NULL
)
go
GRANT INSERT, UPDATE, SELECT ON TAccounts TO PUBLIC
go

CREATE TABLE TActivityTypes
(
	ActivityTypeID	int
		IDENTITY(1,1)
		PRIMARY KEY CLUSTERED,
	ActivityType	varchar(15)	NOT NULL
		UNIQUE NONCLUSTERED,
	Optimist		timestamp	NOT NULL
)
go
GRANT SELECT, UPDATE ON TActivityTypes TO PUBLIC
go

CREATE TABLE TActivity
(
	ActivityID				int
		IDENTITY(1, 1)
		PRIMARY KEY CLUSTERED,
	CustomerID		int		NOT NULL,
	AccountTypeID	int		NOT NULL,
	ActivityTypeID	int		NOT NULL,
	ActivityDateTime	datetime	NOT NULL
		DEFAULT(getdate()),
	ActivityAmount	money		NOT NULL,
	Optimist		timestamp	NOT NULL
)
go
GRANT INSERT, SELECT ON TActivity TO PUBLIC
go

if exists (select * from sysobjects where id = object_id('dbo.Deposit'))
	drop procedure dbo.Deposit
GO
CREATE PROCEDURE Deposit
	@CustomerID int, @AccountType varchar(15), @DepAmount money
AS
SET NOCOUNT ON
	DECLARE @ActivityTypeID int
	DECLARE @AccountTypeID int
	DECLARE @CurrentBalance money
	DECLARE @CustomerName varchar(30)

	IF (@DepAmount < 0)
	BEGIN
		RAISERROR ('Cannot deposit a negative amount.', 16, 1)
		SET NOCOUNT OFF
		RETURN
	END

	SELECT @ActivityTypeID = ActivityTypeID
		FROM TActivityTypes
		WHERE ActivityType = 'Deposit'

	IF @ActivityTypeID IS NULL
	BEGIN
		RAISERROR ('Example database is corrupt, missing ActivityTypes data.', 16, 1)
		SET NOCOUNT OFF
		RETURN
	END

	SELECT @AccountTypeID = AccountTypeID
		FROM TAccountTypes
		WHERE AccountType = @AccountType

	IF @AccountTypeID IS NULL 
	BEGIN
		RAISERROR ('Illegal account type: %s.', 16, 1, @AccountType)
		SET NOCOUNT OFF
		RETURN
	END

	SELECT @CurrentBalance = Balance
		FROM TAccounts
		WHERE CustomerID = @CustomerID AND AccountTypeID = @AccountTypeID

	IF @CurrentBalance IS NULL
	BEGIN
		SELECT @CustomerName = CustomerName FROM TCustomers WHERE CustomerID = @CustomerID
		RAISERROR ('Customer %s does not have a %s account at the bank.', 16, 1, @CustomerName, @AccountType)
		SET NOCOUNT OFF
		RETURN
	END
SET NOCOUNT OFF

	INSERT INTO TActivity (CustomerID, AccountTypeID, ActivityTypeID, ActivityAmount)
		VALUES (@CustomerID, @AccountTypeID, @ActivityTypeID, @DepAmount)
GO
GRANT EXECUTE ON dbo.Deposit TO PUBLIC
GO

if exists (select * from sysobjects where id = object_id('dbo.Withdrawal'))
	drop procedure dbo.Withdrawal
GO
CREATE PROCEDURE Withdrawal
	@CustomerID int, @AccountType varchar(15), @WithdrawAmount money
AS
SET NOCOUNT ON
	DECLARE @ActivityTypeID int
	DECLARE @AccountTypeID int
	DECLARE @CurrentBalance money
	DECLARE @CustomerName varchar(30)

	IF (@WithdrawAmount < 0)
	BEGIN
		RAISERROR ('Cannot withdraw a negative amount.', 16, 1)
		SET NOCOUNT OFF
		RETURN
	END

	SELECT @ActivityTypeID = ActivityTypeID
		FROM TActivityTypes
		WHERE ActivityType = 'Withdrawal'

	IF @ActivityTypeID IS NULL
	BEGIN
		RAISERROR ('Example database is corrupt, missing ActivityTypes data.', 16, 1)
		SET NOCOUNT OFF
		RETURN
	END

	SELECT @AccountTypeID = AccountTypeID
		FROM TAccountTypes
		WHERE AccountType = @AccountType

	IF @AccountTypeID IS NULL 
	BEGIN
		RAISERROR ('Illegal account type: %s.', 16, 1, @AccountType)
		SET NOCOUNT OFF
		RETURN
	END

	SELECT @CurrentBalance = Balance
		FROM TAccounts
		WHERE CustomerID = @CustomerID AND AccountTypeID = @AccountTypeID

	IF @CurrentBalance IS NULL
	BEGIN
		SELECT @CustomerName = CustomerName FROM TCustomers WHERE CustomerID = @CustomerID
		RAISERROR ('Customer %s does not have a %s account at the bank.', 16, 1, @CustomerName, @AccountType)
		SET NOCOUNT OFF
		RETURN
	END

	IF (@CurrentBalance < @WithdrawAmount)
	BEGIN
		SELECT @CustomerName = CustomerName FROM TCustomers WHERE CustomerID = @CustomerID
		RAISERROR ('Customer %s does not have sufficient funds on deposit.', 16, 1, @CustomerName)
		SET NOCOUNT OFF
		RETURN
	END
SET NOCOUNT OFF

	INSERT INTO TActivity (CustomerID, AccountTypeID, ActivityTypeID, ActivityAmount)
		VALUES (@CustomerID, @AccountTypeID, @ActivityTypeID, -(@WithdrawAmount))
GO
GRANT EXECUTE ON dbo.Withdrawal TO PUBLIC
GO

if exists (select * from sysobjects where id = object_id('dbo.AccountCursor'))
	drop procedure dbo.AccountCursor
GO
CREATE PROCEDURE AccountCursor @CustomerID int
AS
SELECT CustomerID, AccountTypeID, Balance
	FROM TAccounts
	WHERE CustomerID = @CustomerID
GO
GRANT EXECUTE ON dbo.AccountCursor TO PUBLIC
GO

if exists (select * from sysobjects where id = object_id('dbo.CustomerHistory'))
	drop procedure dbo.CustomerHistory
GO
CREATE PROCEDURE CustomerHistory @CustomerID int
AS
SELECT ActivityDateTime, AccountType, ActivityType, ActivityAmount
	FROM TActivity A, TAccountTypes AcctT, TActivityTypes ActT
	WHERE A.AccountTypeID = AcctT.AccountTypeID
	AND A.ActivityTypeID = ActT.ActivityTypeID
	AND A.CustomerID = @CustomerID
	ORDER BY ActivityDateTime DESC
GO
GRANT EXECUTE ON dbo.CustomerHistory TO PUBLIC
GO

if exists (select * from sysobjects where id = object_id('dbo.CheckingHistory'))
	drop procedure dbo.CheckingHistory
GO
CREATE PROCEDURE CheckingHistory @CustomerID int
AS
SELECT ActivityDateTime, ActivityType, ActivityAmount
	FROM TActivity A, TAccountTypes AcctT, TActivityTypes ActT
	WHERE A.AccountTypeID = AcctT.AccountTypeID
	AND A.ActivityTypeID = ActT.ActivityTypeID
	AND AcctT.AccountType = 'Checking'
	AND A.CustomerID = @CustomerID
	ORDER BY ActivityDateTime DESC
GO
GRANT EXECUTE ON dbo.CheckingHistory TO PUBLIC
GO

if exists (select * from sysobjects where id = object_id('dbo.SavingsHistory'))
	drop procedure dbo.SavingsHistory
GO
CREATE PROCEDURE SavingsHistory @CustomerID int
AS
SELECT ActivityDateTime, ActivityType, ActivityAmount
	FROM TActivity A, TAccountTypes AcctT, TActivityTypes ActT
	WHERE A.AccountTypeID = AcctT.AccountTypeID
	AND A.ActivityTypeID = ActT.ActivityTypeID
	AND AcctT.AccountType = 'Savings'
	AND A.CustomerID = @CustomerID
	ORDER BY ActivityDateTime DESC
GO
GRANT EXECUTE ON dbo.SavingsHistory TO PUBLIC
GO

if exists (select * from sysobjects where id = object_id('dbo.CustomerList'))
	drop procedure dbo.CustomerList
GO

CREATE PROCEDURE CustomerList AS
SELECT CustomerName, CustomerID
FROM TCustomers
ORDER BY CustomerName
GO

SET IDENTITY_INSERT dbo.TAccountTypes ON
INSERT INTO TAccountTypes (AccountTypeID, AccountType)
	VALUES (1, 'Checking')
INSERT INTO TAccountTypes (AccountTypeID, AccountType)
	VALUES (2, 'Savings')
SET IDENTITY_INSERT dbo.TAccountTypes OFF
GO

SET IDENTITY_INSERT dbo.TActivityTypes ON
INSERT INTO TActivityTypes (ActivityTypeID, ActivityType)
	VALUES (1, 'New Account')
INSERT INTO TActivityTypes (ActivityTypeID, ActivityType)
	VALUES (2, 'Deposit')
INSERT INTO TActivityTypes (ActivityTypeID, ActivityType)
	VALUES (3, 'Withdrawal')
SET IDENTITY_INSERT dbo.TActivityTypes OFF
GO

