// ComboList.cpp : implementation of a ComboList class. The class
//  simplifies filling a combo box control with data from a SQL Server.
//
// This file is part of Microsoft SQL Server online documentation.
// Copyright (C) 1992-1997 Microsoft Corporation. All rights reserved.
//
// This source code is an intended supplement to the Microsoft SQL
// Server online references and related electronic documentation.
#include "stdafx.h"

#include "CmboList.h"
#include "Param.h"
#include "ColAttr.h"
#include "Globals.h"

CComboList::CComboList
    (
    SQLHDBC hdbc,
    int idString,
    CComboBox* pCombo,
    UINT nDispCol
    )
    {
    ASSERT(nDispCol < 4000 /* Max columns in SQL Server SELECT */
        && nDispCol != 0   /* Display column is ordinal, 1 based */);
    ASSERT(pCombo->IsKindOf(RUNTIME_CLASS(CComboBox)));

    m_pBindOffsets  = NULL;
    m_pIndOffsets   = NULL;
    m_cbRow         = 0;
    m_hstmt         = SQL_NULL_HANDLE;
    m_strCommand    = _T("");
    m_nBytesPerBuffer = DEF_BUFFER_CHUNK;

    m_nCols         = 0;

    m_bParamsBound  = FALSE;

    m_pCombo        = pCombo;

    // Internally, we're zero based.
    m_nDispCol      = nDispCol - 1;

    m_hdbc = hdbc;
    if (!m_strCommand.LoadString(idString))
        return;

    if (SQLAllocHandle(SQL_HANDLE_STMT, m_hdbc, &m_hstmt) == SQL_SUCCESS)
        {
        // Set read-only concurrency for this statement, objects of the combo
        //  list class are not updatable.
        SQLSetStmtAttr(m_hstmt, SQL_ATTR_CONCURRENCY,
            (SQLPOINTER) SQL_CONCUR_READ_ONLY, SQL_IS_INTEGER);
        }
    else
        {
        DisplayErrors(SQL_HANDLE_DBC, m_hdbc);
        }
    }

CComboList::~CComboList()
    {
    POSITION    pos;
    CColAttr*   pColAttr;
    PBUFFERCHUNK pBufferChunk;

    if (!m_listColAttr.IsEmpty())
        {
        for (pos = m_listColAttr.GetHeadPosition(); pos != NULL; )
            {
            pColAttr = m_listColAttr.GetNext(pos);
            delete pColAttr;
            }
        }

    if (!m_listBuffers.IsEmpty())
        {
        for (pos = m_listBuffers.GetHeadPosition(); pos != NULL; )
            {
            pBufferChunk = m_listBuffers.GetNext(pos);
            if (pBufferChunk->pBuffer != NULL)
                {
                delete [] pBufferChunk->pBuffer;
                }
            delete pBufferChunk;
            }
        }

    if (m_pBindOffsets != NULL)
        delete [] m_pBindOffsets;

    if (m_pIndOffsets != NULL)
        delete [] m_pIndOffsets;

    SQLFreeHandle(SQL_HANDLE_STMT, m_hstmt);
    }

BOOL CComboList::AddParam
    (
    CParam* pParam
    )
    {
    if (m_hstmt == SQL_NULL_HANDLE)
        {
        return (FALSE);
        }

    m_listParams.AddTail(pParam);
    return (TRUE);
    }

BOOL CComboList::SetBufferSize
    (
    UINT nBytesBuffered
    )
    {
    // No sense in setting this to less than some minimum number of bytes;
    //  we'll use 128. Further, inadvertant huge sizes are trimmed by
    //  reducing the size to the number of bytes set by MAX_BUFFER_CHUNK
    //  in the header.
    ASSERT(nBytesBuffered > MIN_BUFFER_CHUNK);

    nBytesBuffered = min(nBytesBuffered, MAX_BUFFER_CHUNK);

    if (nBytesBuffered > MIN_BUFFER_CHUNK && m_hstmt != SQL_NULL_HANDLE && 
        m_listBuffers.IsEmpty())
        {
        m_nBytesPerBuffer = nBytesBuffered;
        return (TRUE);
        }

    return (FALSE);
    }

BOOL CComboList::Query()
    {
    UINT        nRow;
    UINT        nCol;
    UINT        nBufferChunk;
    PBUFFERCHUNK pBufferChunk;
    PBYTE       pBuffer;
    PTSTR       pValue;
    UINT        nParam;

    if (m_hstmt == SQL_NULL_HANDLE)
        {
        return (FALSE);
        }

    // If we're re-executing, clear the combo box. Any error in re-execution
    //  will leave the box empty.
    if (m_pCombo->GetCount())
        {
        m_pCombo->ResetContent();
        }

    if (m_bParamsBound == FALSE)
        {
        if (m_listParams.GetCount())
            {
            for (nParam = 0; nParam < (UINT) m_listParams.GetCount(); nParam++)
                {
                m_listParams.GetAt(m_listParams.FindIndex(nParam))->Bind(
                    m_hstmt, nParam + 1);
                }
            }
        
        m_bParamsBound = TRUE;
        }

    if (ExecuteStatement() == FALSE)
        {
        return (FALSE);
        }

    // Check to see if we need to determine the shape of the result set.
    if (m_listBuffers.IsEmpty())
        {
        if (m_listColAttr.IsEmpty())
            {
            if (GetColumnAttributes() == FALSE)
                {
                return (FALSE);
                }
            }

        m_pBindOffsets = new UINT[m_nCols];
        m_pIndOffsets  = new UINT[m_nCols];
        if (m_pBindOffsets == NULL || m_pIndOffsets == NULL)
            {
            return (FALSE);
            }

        // GetBindOffsets uses ColAttr column lengths to determine the offsets
        //  for bound values and the total length of a row of bound values
        //  (m_cbRow).  Using "fast re-bind," our application fakes out
        //  ODBC by binding to an address + a couple of bytes. This allows
        //  us to bind to 0 (which would otherwise succeed by doing nothing).
        //  GetBindOffsets adds the faking value to bind and indicator offsets.
        //  We remove it for our use in dereferencing column values after
        //  we're done binding.
        GetBindOffsets();
        m_nRowsPerBuffer = (UINT) (m_nBytesPerBuffer / m_cbRow);

        // Bind the columns.
        if (BindColumns() == FALSE)
            {
            return (FALSE);
            }
        // Re-adjust our offsets. ODBC will not see these again, so we're
        //  safe in doing this.
        for (nCol = 0; nCol < m_nCols; nCol++)
            {
            m_pBindOffsets[nCol] = m_pBindOffsets[nCol] - FAKE_OUT_ODBC;
            m_pIndOffsets[nCol]  = m_pIndOffsets[nCol] - FAKE_OUT_ODBC;
            }

        // Bind the columns, then set the size of our array binding
        //  structure.
        SQLSetStmtAttr(m_hstmt, SQL_ATTR_ROW_BIND_TYPE,
            (SQLPOINTER) m_cbRow, SQL_IS_UINTEGER);
        SQLSetStmtAttr(m_hstmt, SQL_ATTR_ROW_ARRAY_SIZE,
            (SQLPOINTER) m_nRowsPerBuffer, SQL_IS_INTEGER);

        // Set the statement attr for fast re-bind fetch. Once set, the
        //  value in the addressed UINT must be set prior to every fetch 
        //  or the pointer itself must be re-set to NULL. If either of
        //  those is not done, application behavior is undefined.
        SQLSetStmtAttr(m_hstmt, SQL_ATTR_ROW_BIND_OFFSET_PTR,
            (SQLPOINTER) &m_pBufferAsUint, SQL_IS_POINTER);
        }

    if (FetchRecords() == FALSE)
        {
        return (FALSE);
        }

    // Add strings to the combo box.
    for (nBufferChunk = 0; nBufferChunk < (UINT) m_listBuffers.GetCount();
        nBufferChunk++)
        {
        pBufferChunk = m_listBuffers.GetAt(m_listBuffers.FindIndex(nBufferChunk));
        pBuffer = pBufferChunk->pBuffer;
        for (nRow = 0; nRow < pBufferChunk->nRowsInBuffer; nRow++)
            {
            pValue = (PTSTR) (pBuffer + (m_cbRow * nRow) + 
                m_pBindOffsets[m_nDispCol]);
            m_pCombo->AddString(pValue);
            }
        }

    return (TRUE);
    }

BOOL CComboList::GetValue
    (
    UINT nCol,
    PVOID* ppValue,
    BOOL* pbIsNULL
    )
    {
    PBYTE       pValue;
    PBYTE       pBuffer;
    UINT        nRow;

    // Use ASSERT to trap these errors at development. Run-time checks are
    //  always performed.
    ASSERT(nCol != 0 && nCol <= m_nCols);
    ASSERT(ppValue != NULL && pbIsNULL != NULL);

    if (nCol == 0 || nCol > m_nCols)
        {
        return (FALSE);
        }
    if (ppValue == NULL || pbIsNULL == NULL)
        {
        return (FALSE);
        }

    // Set for zero-based processing internally.
    nCol--;

    nRow = (UINT) (m_pCombo->GetCurSel());
    if (nRow == (UINT) CB_ERR)
        {
        return (FALSE);
        }

    pBuffer 
        = m_listBuffers.GetAt(
            m_listBuffers.FindIndex((int) (nRow / m_nRowsPerBuffer)))->pBuffer;
    if (pBuffer == NULL)
        {
        return (FALSE);
        }

    if (*((SQLINTEGER*) (pBuffer + (m_cbRow * (nRow % m_nRowsPerBuffer)) +
        m_pIndOffsets[nCol])) == SQL_NULL_DATA)
        {
        *pbIsNULL = TRUE;
        return (TRUE);
        }

    *pbIsNULL = FALSE;
    pValue = (pBuffer + (m_cbRow * (nRow % m_nRowsPerBuffer)) + 
        m_pBindOffsets[nCol]);
    *ppValue = pValue;

    return (TRUE);
    }

BOOL CComboList::ExecuteStatement()
    {
    // Set up statement handle. Get a forward only, non-blocking 
    //  statement by adjusting the buffer size. We use a default
    //  value here. Before fetching, we'll adjust this to the 
    //  actual number of rows that will fit in each buffer.
    SQLSetStmtAttr(m_hstmt, SQL_ATTR_ROW_ARRAY_SIZE, (SQLPOINTER) 2,
        SQL_IS_INTEGER);

    // Use direct execution; developer can optimize by using a SQL Server
    //  stored procedure.
    if (SQL_SUCCEEDED(SQLExecDirect(m_hstmt,
        (SQLTCHAR*) ((LPCTSTR) m_strCommand), SQL_NTS)))
        {
        return (TRUE);
        }

    DisplayErrors(SQL_HANDLE_STMT, m_hstmt);
    return (FALSE);
    }

BOOL CComboList::GetColumnAttributes()
    {
    SQLSMALLINT nCols;
    SQLSMALLINT nCol;
    CColAttr*   pColAttr;

    if (SQLNumResultCols(m_hstmt, &nCols) != SQL_SUCCESS)
        {
        DisplayErrors(SQL_HANDLE_STMT, m_hstmt);
        return (FALSE);
        }

    if (nCols == 0)
        {
        return (FALSE);
        }

    if (m_nDispCol >= (UINT) nCols)
        {
        return (FALSE);
        }

    for (nCol = 0; nCol < nCols; nCol++)
        {
        pColAttr = new CColAttr(m_hstmt, nCol + 1);
        if (SQL_SUCCEEDED(pColAttr->GetAttr()))
            {
            m_listColAttr.AddTail(pColAttr);
            }
        else
            {
            return (FALSE);
            }
        }

    m_nCols = (UINT) nCols;
    return (TRUE);
    }

void CComboList::GetBindOffsets()
    {
    UINT        nCol;
    CColAttr*   pColAttr;

    m_cbRow = 0;

    for (nCol = 0; nCol < m_nCols; nCol++)
        {
        pColAttr = m_listColAttr.GetAt(m_listColAttr.FindIndex(nCol));

        m_pBindOffsets[nCol] = m_cbRow + FAKE_OUT_ODBC;
        m_cbRow += pColAttr->GetBoundBufferWidth();

        m_pIndOffsets[nCol] = m_cbRow + FAKE_OUT_ODBC;
        m_cbRow += sizeof(SQLINTEGER);
        }
    }

// BindColumns indicates only an offset to the data. Objects of the
//  ComboList class use the fast re-bind method implemented in
//  ODBC 3.0. Fast re-binding indicates an offset for binding and a
//  SQLPOINTER for the address against which to apply the bound
//  offsets.
BOOL CComboList::BindColumns()
    {
    UINT        nCol;
    SQLSMALLINT eCType;
    SQLINTEGER  cbBuffer;
    CColAttr*   pColAttr;

    for (nCol = 0; nCol < m_nCols; nCol++)
        {
        pColAttr = m_listColAttr.GetAt(m_listColAttr.FindIndex(nCol));

        // Bind the display column to char no matter its base type,
        //  allowing the SQL Server driver to handle conversion for us.
        if (nCol == m_nDispCol)
            {
            eCType = SQL_C_TCHAR;
            cbBuffer = (SQLINTEGER) pColAttr->GetCharBufferWidth();
            }
        else
            {
            eCType = pColAttr->GetDefaultBindType();
            cbBuffer = (SQLINTEGER) pColAttr->GetBoundBufferWidth();
            }

        if (!SQL_SUCCEEDED(SQLBindCol(m_hstmt, nCol + 1, eCType,
            (SQLPOINTER) (m_pBindOffsets[nCol]), cbBuffer,
            (SQLINTEGER*) (m_pIndOffsets[nCol]))))
            {
            DisplayErrors(SQL_HANDLE_STMT, m_hstmt);
            return (FALSE);
            }
        }

    return (TRUE);
    }

// FetchRecords for the ComboList class illustrates rapid rebinding by
//  adjusting the 
BOOL CComboList::FetchRecords()
    {
    UINT        nBuffer = 0;
    PBUFFERCHUNK pBufferChunk;
    PBYTE       pBuffer;
    POSITION    pos;
    
    m_nRowsTotal = 0;
    while (m_nRowsTotal < ROW_LIMIT)
        {
        pos = m_listBuffers.FindIndex(nBuffer);
        if (pos != NULL)
            {
            pBufferChunk = m_listBuffers.GetAt(pos);
            pBuffer = pBufferChunk->pBuffer;
            }
        else
            {
            pBufferChunk = new BUFFERCHUNK;
            if (pBufferChunk == NULL)
                {
                return (FALSE);
                }

            pBuffer = new BYTE[m_nBytesPerBuffer];
            if (pBuffer == NULL)
                {
                delete pBufferChunk;
                return (FALSE);
                }
            pBufferChunk->pBuffer = pBuffer;
            m_listBuffers.AddTail(pBufferChunk);
            }
    
        // Set statement attribute to capture number of rows fetched.
        SQLSetStmtAttr(m_hstmt, SQL_ATTR_ROWS_FETCHED_PTR,
            (SQLPOINTER) &(pBufferChunk->nRowsInBuffer), SQL_IS_POINTER);

        // Set the address for the fast re-bind fetch.
        m_pBufferAsUint = (SQLUINTEGER) (pBuffer - FAKE_OUT_ODBC);

        // In ODBC 3.x, SQLFetch supports arrays of bound rows or columns.
        //  SQLFetchScroll (or ODBC 2.x SQLExtendedFetch) is not necessary
        //  to support fastest retrieval of our data rows.
        switch (SQLFetch(m_hstmt))
            {
            case SQL_SUCCESS:
            case SQL_SUCCESS_WITH_INFO:
                {
                // We test to see if it was the last fetch (number of
                //  rows fetched less than the number that will fit
                //  in a buffer). If so, we close the cursor and return,
                //  else we keep fetching additional buffers.
                m_nRowsTotal += pBufferChunk->nRowsInBuffer;

                if (pBufferChunk->nRowsInBuffer < m_nRowsPerBuffer)
                    {
                    SQLEndTran(SQL_HANDLE_DBC, m_hdbc, SQL_ROLLBACK);
                    SQLCloseCursor(m_hstmt);

                    return (TRUE);
                    }
                break;
                }

            case SQL_NO_DATA:
                {
                // We hit here if the buffer size holds an exact multiple
                //  of our row size (for example, the SQL statement returns
                //  no rows at all). Delete the last created buffer so we
                //  don't waste memory needlessly and return.
                m_listBuffers.RemoveTail();

                delete [] pBuffer;
                delete pBufferChunk;

                return (TRUE);
                }

            default:
                {
                DisplayErrors(SQL_HANDLE_STMT, m_hstmt);
                return (FALSE);
                }
            }
        
        nBuffer++;
        }

    // More rows are available, but return TRUE to display what's been
    //  retrieved.
    SQLEndTran(SQL_HANDLE_DBC, m_hdbc, SQL_ROLLBACK);
    SQLCloseCursor(m_hstmt);

    return (TRUE);
    }

