// ComboList.h : header file for the ComboList class.
//
// This file is part of Microsoft SQL Server online documentation.
// Copyright (C) 1992-1997 Microsoft Corporation. All rights reserved.
//
// This source code is an intended supplement to the Microsoft SQL
// Server online references and related electronic documentation.
#ifndef __COMBOLIST_H__
#define __COMBOLIST_H__

#ifndef __AFXWIN_H__
	#error include 'stdafx.h' before including this file for PCH
#endif

#include "afxtempl.h"
#include "SQL.h"
#include "SQLExt.h"

// Macros
#define DEF_BUFFER_CHUNK        4096
#define MIN_BUFFER_CHUNK        128
#define MAX_BUFFER_CHUNK        65536
#define ROW_LIMIT               1024
#define FAKE_OUT_ODBC           4

// Types
typedef struct tagBUFFERCHUNK
    {
    PBYTE       pBuffer;
    UINT        nRowsInBuffer;
    } BUFFERCHUNK;
typedef BUFFERCHUNK* PBUFFERCHUNK;

// Forward declarations
class CParam;
class CColAttr;

class CComboList
{
public:
    // Construction requires 1) a database connection; 2) the identifier of a 
    //  string resource containing an ODBC or Transact-SQL SQL statement; 3)
    //  a pointer to an MFC ComboBox control; and 4) the ordinal of the column
    //  that will be displayed in the ComboBox for user selection.
    CComboList(SQLHDBC hdbc, int idString, CComboBox* pCombo, UINT nDispCol);

    // SetBufferSize adjusts the size of the rows-fetched buffer. Must be
    //  called prior to initial execution (first call to member function
    //  Query()). The buffer size is "blocked." Additional blocks of the
    //  size indicated will be allocated to fetch all rows if possible. A
    //  maximimum of 1024 rows are retrieved regardless of the buffer size.
    BOOL SetBufferSize(UINT nBytesBuffered);

    // AddParam adds the passed parameter to a list of parameters available
    //  to statement execution. Parameters are owned by the caller. The
    //  caller must ensure that the parameter does not go out of scope before
    //  the ComboList does and must delete the parameter at an appropriate
    //  time.
    BOOL AddParam(CParam* pParam);

    // Query executes the SQL statement associated with the list and fills
    //  the ComboBox control with the fetched rows.
    BOOL Query();

    // GetValue returns a pointer to the value requested from the ComboList.
    //  The list maintains multiple data columns for a single displayed
    //  string value. Data values are retrieved by their row and column
    //  indices. The row is the current selection of the combo box. The 
    //  column index is one based. If a combo list is created from an ODBC
    //  or Transact-SQL statement that generates 3 result columns and the
    //  3rd column is displayed, the data columns are referenced with 1 and
    //  2, not 0 and 1.
    //  
    // If a value exists, the value of ppValue is set to a pointer to the
    //  data and the value of pbIsNULL is set to FALSE. The function returns
    //  TRUE.
    // If the row and column received are valid, but the value is null, ppValue
    //  is set to NULL and pbIsNULL is set to TRUE. The function returns TRUE.
    // If nCol is invalid (0 or greater than the available number of columns),
    //   or the combo box's GetCurSel() function returns CB_ERR, or ppValue or
    //   pbIsNull are themselves NULL, the function returns FALSE.
    //
    // At all times, the caller is responsible for knowing what data type the
    //  value has.
    BOOL GetValue(UINT nCol, PVOID* ppValue, BOOL* pbIsNULL);

    virtual ~CComboList();

private:
    // The hdbc, ODBC hstmt and Transact or ODBC SQL statement.
    SQLHDBC     m_hdbc;
    SQLHSTMT    m_hstmt;
    CString     m_strCommand;

    // ComboBox data, the ComboBox will be used to display rows from a single
    //  SQL Server column.
    CComboBox*  m_pCombo;
    UINT        m_nDispCol;

    // A list of parameters for the ODBC statement.
    CList<CParam*, CParam*>     m_listParams;
    BOOL        m_bParamsBound;

    // A list of ColumnAttributes from the executed statement.
    CList<CColAttr*, CColAttr*> m_listColAttr;

    // Data and offsets to bound values.
    UINT        m_nBytesPerBuffer;
    UINT        m_nRowsPerBuffer;
    PUINT       m_pBindOffsets;
    PUINT       m_pIndOffsets;
    UINT        m_cbRow;
    CList<PBUFFERCHUNK, PBUFFERCHUNK> m_listBuffers;
    SQLUINTEGER m_pBufferAsUint;

    // Number of records and columns in result set.
    UINT        m_nRowsTotal;
    UINT        m_nCols;

private:
    BOOL ExecuteStatement();
    BOOL GetColumnAttributes();
    void GetBindOffsets();
    BOOL BindColumns();
    BOOL FetchRecords();
};

#endif  // __COMBOLIST_H__
