// Cursor.cpp
//
// This file is part of Microsoft SQL Server online documentation.
// Copyright (C) 1992-1997 Microsoft Corporation. All rights reserved.
//
// This source code is an intended supplement to the Microsoft SQL
// Server online references and related electronic documentation.
#include "stdafx.h"

#include "Cursor.h"
#include "ColAttr.h"
#include "Param.h"
#include "Globals.h"

CCursor::CCursor
    (
    SQLHDBC hdbc,
    LPCTSTR strOpenCursor,
    SQLINTEGER eCursorType,
    SQLINTEGER eConcur)
    {
    // Assert on debug construction if required parameters are invalid.
    ASSERT(hdbc != SQL_NULL_HANDLE && strOpenCursor != NULL &&
        _tcslen(strOpenCursor));

    m_nRowsetSize       = 1;
    m_nRows             = 0;
    m_nCols             = 0;
    m_nCurrentRow       = 0;
    m_bDataBound        = FALSE;
    m_bparamsBound      = FALSE;
    m_pRowData          = NULL;
    m_pBindOffsets      = NULL;
    m_pIndOffsets       = NULL;
    m_hstmt             = SQL_NULL_HANDLE;
    m_hdbc              = SQL_NULL_HANDLE;
    m_nTimeout          = TIMEOUT_MAX;
    m_strUpdate         = _T("");
    m_strDelete         = _T("");
    m_strCursorName     = _T("");
    
    // Record parameters...
    m_strOpen = strOpenCursor;
    
    // ... for now, we don't allow overriding the default cursor type.
    m_eCursorType = SQL_CURSOR_KEYSET_DRIVEN;
    m_eConcur = eConcur;

    if (!SQL_SUCCEEDED(SQLAllocHandle(SQL_HANDLE_STMT, hdbc,
        &m_hstmt)))
        {
        DisplayErrors(SQL_HANDLE_DBC, m_hdbc);
        return;
        }

    // Get cursor of type with the concurrency control level we desire, then
    //  set the timeout on statement execution. Because we use the server to
    //  handle concurrency, we need to ensure that we don't execute any
    //  statements on this handle that cannot timeout.
    if (
        !SQL_SUCCEEDED(SQLSetStmtAttr(m_hstmt, SQL_ATTR_CURSOR_TYPE,
            (SQLPOINTER) m_eCursorType, SQL_IS_INTEGER))    ||
        !SQL_SUCCEEDED(SQLSetStmtAttr(m_hstmt, SQL_ATTR_CONCURRENCY,
            (SQLPOINTER) m_eConcur, SQL_IS_INTEGER))        ||
        !SQL_SUCCEEDED(SQLSetStmtAttr(m_hstmt, SQL_ATTR_QUERY_TIMEOUT,
            (SQLPOINTER) m_nTimeout, SQL_IS_INTEGER)))
        {
        DisplayErrors(SQL_HANDLE_STMT, m_hstmt);
        SQLFreeHandle(SQL_HANDLE_STMT, m_hstmt);
        m_hstmt = NULL;
        return;
        }

    // We only get to record this if everything succeeds values.
    m_hdbc    = hdbc;
    }

CCursor::~CCursor()
    {
    if (m_hstmt != SQL_NULL_HANDLE)
        {
        SQLCloseCursor(m_hstmt);
        SQLFreeHandle(SQL_HANDLE_STMT, m_hstmt);

        if (m_pBindOffsets != NULL)
            {
            delete [] m_pBindOffsets;
            }

        if (m_pIndOffsets != NULL)
            {
            delete [] m_pIndOffsets;
            }

        if (m_listColAttr.GetCount())
            {
            SQLSMALLINT nCol;
            for (nCol = 0; nCol < m_nCols; nCol++)
                {
                delete m_listColAttr.GetAt(m_listColAttr.FindIndex(nCol));
                }
            }
        }
    }

BOOL CCursor::AddParam
    (
    CParam* pParam
    )
    {
    if (m_hstmt == SQL_NULL_HANDLE)
        {
        return (FALSE);
        }

    m_listParams.AddTail(pParam);
    return (TRUE);
    }

SQLRETURN CCursor::SetRowArraySize
    (
    UINT nRows
    )
    {
    ASSERT(nRows > 0 && nRows <= ROWARRAY_MAX);
    if (nRows == 0 || nRows > ROWARRAY_MAX)
        {
        return (SQL_ERROR);
        }

    m_nRowsetSize = nRows;
    return (SQL_SUCCESS);
    }

BOOL CCursor::SetDataPtrs
    (
    PUINT pBindOffsets,
    PUINT pIndOffsets,
    PBYTE pRowData
    )
    {
    ASSERT(pBindOffsets && pIndOffsets && pRowData);
    if (pBindOffsets == NULL || pIndOffsets == NULL || pRowData == NULL)
        {
        return (FALSE);
        }

    m_pClientBindOffsets = pBindOffsets;
    m_pClientIndOffsets = pIndOffsets;
    m_pRowData = pRowData;

    return (TRUE);
    }

SQLRETURN CCursor::Delete()
    {
    SQLRETURN   sRet;

    // Assert in debug build, check anyway for run time.
    ASSERT(m_nCurrentRow != 0);
    if (m_nCurrentRow == 0)
        {
        return (SQL_ERROR);
        }

    while (TRUE)
        {
        sRet = SQLSetPos(m_hstmt, m_nCurrentRow, SQL_DELETE,
            SQL_LOCK_NO_CHANGE);

        if (sRet == SQL_SUCCESS_WITH_INFO)
            {
            SQLINTEGER  nNativeError = 0;

            // Check to see if the native error is SQL Server 16934,
            //  optimistic concurrency collision.
            SQLGetDiagField(SQL_HANDLE_STMT, m_hstmt, 1, SQL_DIAG_NATIVE,
                (SQLPOINTER) &nNativeError, SQL_IS_INTEGER, NULL);

            // If it's a status we don't know about, display the message,
            //  then act as if successful.
            if (nNativeError == 16934)
                {
                if (!SQL_SUCCEEDED(SQLSetPos(m_hstmt, m_nCurrentRow,
                    SQL_REFRESH, SQL_LOCK_NO_CHANGE)))
                    {
                    DisplayErrors(SQL_HANDLE_STMT, m_hstmt);
                    return (SQL_ERROR);
                    }
                }
            else
                {
                DisplayErrors(SQL_HANDLE_STMT, m_hstmt);
                break;
                }
            }
        else
            {
            // Break out of loop on !SQL_SUCCESS_WITH_INFO.
            if (sRet != SQL_SUCCESS)
                {
                DisplayErrors(SQL_HANDLE_STMT, m_hstmt);
                }
            break;
            }
        }

    return (sRet);
    }

SQLRETURN CCursor::Query()
    {
    SQLSMALLINT nCol;
    SQLSMALLINT nParam;
    CColAttr*   pColAttr;
    SQLRETURN   sRet;

    // Use assert for debugging, run time check anyway.
    ASSERT(m_hstmt != SQL_NULL_HANDLE);
    if (m_hstmt == SQL_NULL_HANDLE)
        {
        return (SQL_ERROR);
        }

    // If we're "re-querying", close the cursor.
    if (m_nCols != 0)
        {
        SQLCloseCursor(m_hstmt);
        }

    // Test for bound parameters and bind them if necessary.
    if (m_bparamsBound == FALSE)
        {
        if (m_listParams.GetCount())
            {
            for (nParam = 0; nParam < m_listParams.GetCount(); nParam++)
                {
                (m_listParams.GetAt(
                    m_listParams.FindIndex(nParam)))->Bind(
                    m_hstmt, nParam + 1);
                }
            }
        
        m_bparamsBound = TRUE;
        }

    // Execute our cursor generating SQL statement. Prepared execution
    //  makes no sense for cursor generating statement handles.
    if (!SQL_SUCCEEDED(sRet = SQLExecDirect(m_hstmt,
        (SQLTCHAR*) ((LPCTSTR) m_strOpen), SQL_NTS)))
        {
        DisplayErrors(SQL_HANDLE_STMT, m_hstmt);
        return (sRet);
        }
        
    // If first time through, then we need to collect column attributes.
    if (m_nCols == 0)
        {
        if (!SQL_SUCCEEDED(sRet = SQLNumResultCols(m_hstmt, &m_nCols)))
            {
            DisplayErrors(SQL_HANDLE_STMT, m_hstmt);
            return (sRet);
            }

        if (m_nCols == 0)
            {
            return (SQL_ERROR);
            }

        for (nCol = 0; nCol < m_nCols; nCol++)
            {
            pColAttr = new CColAttr(m_hstmt, nCol+1);
            if (SQL_SUCCEEDED(pColAttr->GetAttr()))
                {
                m_listColAttr.AddTail(pColAttr);
                }
            else
                {
                DisplayErrors(SQL_HANDLE_STMT, m_hstmt);
                return (SQL_ERROR);
                }
            }
        
        m_pBindOffsets = new UINT[m_nCols];
        m_pIndOffsets = new UINT[m_nCols];

        GetBindOffsets();
        }

    return (sRet);
    }

SQLRETURN CCursor::Fetch()
    {
    SQLRETURN   sRet;

    // Use assert for debugging, run time check anyway.
    ASSERT(m_hstmt != SQL_NULL_HANDLE && m_pRowData != NULL);
    if (m_hstmt == SQL_NULL_HANDLE || m_pRowData == NULL)
        {
        return (SQL_ERROR);
        }

    // Set the row fetch size every time. The client can alter the buffer
    //  (cursor block) size.
    if (!SQL_SUCCEEDED(sRet = SQLSetStmtAttr(m_hstmt, SQL_ATTR_ROW_ARRAY_SIZE,
        (SQLPOINTER) m_nRowsetSize, SQL_IS_INTEGER)))
        {
        DisplayErrors(SQL_HANDLE_STMT, m_hstmt);
        return (sRet);
        }

    if (m_bDataBound == FALSE)
        {
        memcpy(m_pClientBindOffsets, m_pBindOffsets, sizeof(UINT) * m_nCols);
        memcpy(m_pClientIndOffsets, m_pIndOffsets, sizeof(UINT) * m_nCols);

        if ((m_bDataBound = BindColumns()) == FALSE)
            {
            return (SQL_ERROR);
            }

        SQLSetStmtAttr(m_hstmt, SQL_ATTR_ROW_BIND_TYPE, (SQLPOINTER) m_cbRow,
            SQL_IS_UINTEGER);

        SQLSetStmtAttr(m_hstmt, SQL_ATTR_ROWS_FETCHED_PTR,
            (SQLPOINTER) &m_nRows, SQL_IS_POINTER);
        }

    // SQLFetch will time out and return an error on lock collision.
    if (SQL_SUCCEEDED(sRet = SQLFetchScroll(m_hstmt, SQL_FETCH_NEXT, 0)))
        {
        m_nCurrentRow = 1;
        }
    else
        {
        DisplayErrors(SQL_HANDLE_STMT, m_hstmt);
        m_nCurrentRow = 0;
        }

    return (sRet);
    }

LPCTSTR CCursor::GetCursorName()
    {
    // Only get the cursor name from the server if we're asked for it.
    if (m_nCols != 0 && m_strCursorName.GetLength() == 0)
        {
        PTSTR pCursorName = 
            m_strCursorName.GetBufferSetLength(CURSORNAME_MAX + 1);
        SQLSMALLINT cbCursorName;

        if (SQLGetCursorName(m_hstmt, (SQLTCHAR*) pCursorName,
            CURSORNAME_MAX + 1, &cbCursorName) != SQL_SUCCESS)
            {
            _tcscpy(pCursorName, _T(""));
            }

        m_strCursorName.ReleaseBuffer();
        }

    // Will return a zero length string unless we've gotten a name from
    //  SQL Server.
    return ((LPCTSTR) m_strCursorName);
    }

SQLRETURN CCursor::Position
    (
    UINT nRow
    )
    {
    SQLRETURN   sRet;

    // Assert in debug build, check anyway for run time.
    ASSERT(m_nRows != 0 && nRow != 0 && nRow <= m_nRows);
    if (m_nRows == 0 || nRow == 0 || nRow > m_nRows)
        {
        return (SQL_ERROR);
        }

    if (SQL_SUCCEEDED(sRet = SQLSetPos(m_hstmt, nRow, SQL_POSITION,
        SQL_LOCK_NO_CHANGE)))
        {
        m_nCurrentRow = nRow;
        }

    return (sRet);
    }

SQLRETURN CCursor::Update()
    {
    SQLRETURN   sRet;

    // Assert in debug build, check anyway for run time.
    ASSERT(m_nCurrentRow != 0);
    if (m_nCurrentRow == 0)
        {
        return (SQL_ERROR);
        }

    while (TRUE)
        {
        switch (sRet = SQLSetPos(m_hstmt, m_nCurrentRow, SQL_UPDATE,
            SQL_LOCK_NO_CHANGE))
            {
            case (SQL_SUCCESS_WITH_INFO):
                {
                TCHAR   acSQLState[SQL_SQLSTATE_SIZE + 1];
                SQLSMALLINT cbSQLState;

                // Check the SQLSTATE on SUCCESS_WITH_INFO. Although the ODBC
                //  documents state basing processing logic on SQLSTATE is
                //  problematic, the SQL Server driver will consistently
                //  return state 01001 when an optimistic concurrency
                //  UPDATE statement fails.
                SQLGetDiagField(SQL_HANDLE_STMT, m_hstmt, 1,
                    SQL_DIAG_SQLSTATE, (SQLPOINTER) acSQLState, 
                    SQL_SQLSTATE_SIZE + 1, &cbSQLState);

                // If it's not the state we expected, display the message,
                //  then act as if successful.
                if (_tcscmp(acSQLState, _T("01001")) != 0)
                    {
                    DisplayErrors(SQL_HANDLE_STMT, m_hstmt);
                    sRet = SQL_SUCCESS;
                    }
                }

            case (SQL_SUCCESS):
                {
                if (!SQL_SUCCEEDED(SQLSetPos(m_hstmt, m_nCurrentRow, SQL_REFRESH,
                    SQL_LOCK_NO_CHANGE)))
                    {
                    DisplayErrors(SQL_HANDLE_STMT, m_hstmt);
                    return (SQL_ERROR);
                    }
                break;
                }

            default:
                {
                DisplayErrors(SQL_HANDLE_STMT, m_hstmt);
                break;
                }
            }

        // Break out of loop on !SQL_SUCCESS_WITH_INFO.
        if (sRet != SQL_SUCCESS_WITH_INFO)
            {
            break;
            }
        }

    return (sRet);
    }

SQLRETURN CCursor::Update
    (
    LPCTSTR strPositionedUpdate
    )
    {
    SQLRETURN   sRet;
    SQLHSTMT    hstmt;

    // Assert in debug build, check anyway for run time.
    ASSERT(m_nCurrentRow != 0 && m_strCursorName.GetLength());
    if (m_nCurrentRow == 0 || m_strCursorName.GetLength() == 0)
        {
        return (SQL_ERROR);
        }

    if ((_tcsncmp(strPositionedUpdate, _T("UPDATE"), 6) != 0 &&
        _tcsncmp(strPositionedUpdate, _T("DELETE"), 6) != 0) ||
        _tcsstr(strPositionedUpdate, (LPCTSTR) m_strCursorName) == NULL)
        {
        return (SQL_ERROR);
        }

    if (!SQL_SUCCEEDED(SQLAllocHandle(SQL_HANDLE_STMT, m_hdbc, &hstmt)))
        {
        DisplayErrors(SQL_HANDLE_DBC, m_hdbc);
        return (SQL_ERROR);
        }

    while (TRUE)
        {
        if (!SQL_SUCCEEDED(sRet = SQLSetPos(m_hstmt, m_nCurrentRow,
            SQL_POSITION, SQL_LOCK_NO_CHANGE)))
            {
            DisplayErrors(SQL_HANDLE_DBC, m_hdbc);
            break;
            }

        switch (sRet = SQLExecDirect(hstmt,
            (SQLTCHAR*) strPositionedUpdate, SQL_NTS))
            {
            case (SQL_SUCCESS_WITH_INFO):
                {
                TCHAR   acSQLState[SQL_SQLSTATE_SIZE + 1];
                SQLSMALLINT cbSQLState;

                // Check the SQLSTATE on SUCCESS_WITH_INFO. Although the ODBC
                //  documents state basing processing logic on SQLSTATE is
                //  problematic, the SQL Server driver will consistently
                //  return state 01001 when an optimistic concurrency
                //  UPDATE statement fails.
                SQLGetDiagField(SQL_HANDLE_STMT, hstmt, 1,
                    SQL_DIAG_SQLSTATE, (SQLPOINTER) acSQLState, 
                    SQL_SQLSTATE_SIZE + 1, &cbSQLState);

                // If it's not the state we expected, display the message,
                //  then act as if successful.
                if (_tcscmp(acSQLState, _T("01001")) == 0)
                    {
                    if (!SQL_SUCCEEDED(SQLSetPos(m_hstmt, m_nCurrentRow,
                        SQL_REFRESH, SQL_LOCK_NO_CHANGE)))
                        {
                        DisplayErrors(SQL_HANDLE_STMT, m_hstmt);
                        sRet = SQL_ERROR;
                        }
                    }
                else
                    {
                    DisplayErrors(SQL_HANDLE_STMT, hstmt);
                    }

                break;
                }

            case (SQL_SUCCESS):
                {
                SQLINTEGER  nRowsAff;
                SQLRowCount(hstmt, &nRowsAff);

                if (!SQL_SUCCEEDED(SQLSetPos(m_hstmt, m_nCurrentRow,
                    SQL_REFRESH, SQL_LOCK_NO_CHANGE)))
                    {
                    DisplayErrors(SQL_HANDLE_STMT, m_hstmt);
                    sRet = SQL_ERROR;
                    }
                break;
                }

            default:
                {
                DisplayErrors(SQL_HANDLE_STMT, hstmt);
                break;
                }
            }

        // If we failed update because of optimistic concurrency failure,
        //  repeat the loop.
        if (sRet != SQL_SUCCESS_WITH_INFO)
            {
            break;
            }
        }

    SQLFreeHandle(SQL_HANDLE_STMT, hstmt);
    return (sRet);
    }

void CCursor::GetBindOffsets()
    {
    UINT        nCol;

    m_cbRow = 0;

    for (nCol = 0; nCol < (UINT) m_nCols; nCol++)
        {
        m_pBindOffsets[nCol] = m_cbRow;
        m_cbRow += 
            (m_listColAttr.GetAt(
            m_listColAttr.FindIndex(nCol)))->GetBoundBufferWidth();
        m_pIndOffsets[nCol] = m_cbRow;
        m_cbRow += sizeof(SQL_INTEGER);
        }
    }

// Bind columns, then set statement to bind by row for array fetches.
BOOL CCursor::BindColumns()
    {
    UINT        nCol;
    CColAttr*   pColAttr;

    for (nCol = 0; nCol < (UINT) m_listColAttr.GetCount(); nCol++)
        {
        pColAttr = m_listColAttr.GetAt(m_listColAttr.FindIndex(nCol));

        if (!SQL_SUCCEEDED(SQLBindCol(m_hstmt, nCol + 1,
            pColAttr->GetDefaultBindType(),
            (SQLPOINTER) (m_pRowData + m_pBindOffsets[nCol]),
            (SQLINTEGER) pColAttr->GetBoundBufferWidth(),
            (SQLINTEGER*) (m_pRowData + m_pIndOffsets[nCol]))))
            {
            DisplayErrors(SQL_HANDLE_STMT, m_hstmt);
            return (FALSE);
            }
        }

    return (TRUE);
    }

