// LstCList.cpp
//
// This file is part of Microsoft SQL Server online documentation.
// Copyright (C) 1992-1997 Microsoft Corporation. All rights reserved.
//
// This source code is an intended supplement to the Microsoft SQL
// Server online references and related electronic documentation.
#include "stdafx.h"

#include "LstCList.h"
#include "Param.h"
#include "ColAttr.h"
#include "Globals.h"

CListCtlList::CListCtlList
    (
    SQLHDBC hdbc,
    int idString,
    CListCtrl* pList
    )
    {
    ASSERT(pList->IsKindOf(RUNTIME_CLASS(CListCtrl)));

    m_pRecords      = NULL;
    m_pBindOffsets  = NULL;
    m_pIndOffsets   = NULL;
    m_cbRow         = 0;
    m_hstmt         = SQL_NULL_HANDLE;
    m_strCommand    = _T("");
    m_bBlocking     = TRUE;
    m_nRowsBuffered = MIN_BUFFERED_ROWS;

    m_nRowsFetched  = 0;
    m_nCols         = 0;

    m_bParamsBound  = FALSE;

    m_pList         = pList;
    m_nGridRows     = MIN_GRID_ROWS;
    m_nRowsInFirstPage = 0;

    m_hdbc = hdbc;
    if (!m_strCommand.LoadString(idString))
        return;

    if (SQLAllocHandle(SQL_HANDLE_STMT, m_hdbc, &m_hstmt) == SQL_SUCCESS)
        {
        // Set read-only concurrency for this statement; the grid list
        //  class is not updatable.
        SQLSetStmtAttr(m_hstmt, SQL_ATTR_CONCURRENCY,
            (SQLPOINTER) SQL_CONCUR_READ_ONLY, SQL_IS_INTEGER);

        // Set statement attribute to capture number of rows fetched.
        SQLSetStmtAttr(m_hstmt, SQL_ATTR_ROWS_FETCHED_PTR,
            (SQLPOINTER) &m_nRowsFetched, SQL_IS_POINTER);
        }
    else
        {
        DisplayErrors(SQL_HANDLE_DBC, m_hdbc);
        }
    }

CListCtlList::~CListCtlList()
    {
    CColAttr*   pColAttr;
    POSITION    pos;

    if (!m_listColAttr.IsEmpty())
        {
        for (pos = m_listColAttr.GetHeadPosition(); pos != NULL; )
            {
            pColAttr = m_listColAttr.GetNext(pos);
            delete pColAttr;
            }
        }

    if (m_pRecords != NULL)
        delete [] m_pRecords;

    if (m_pBindOffsets != NULL)
        delete [] m_pBindOffsets;

    if (m_pIndOffsets != NULL)
        delete [] m_pIndOffsets;

    SQLFreeHandle(SQL_HANDLE_STMT, m_hstmt);
    }

BOOL CListCtlList::AddParam
    (
    CParam* pParam
    )
    {
    if (m_hstmt == SQL_NULL_HANDLE)
        {
        return (FALSE);
        }

    m_listParams.AddTail(pParam);
    return (TRUE);
    }

BOOL CListCtlList::SetBufferSize
    (
    UINT nRowsBuffered
    )
    {
    // No sense in setting this to one or less; need one row for the column
    //  headings and at least one row of data.
    ASSERT(nRowsBuffered > 1);

    if (nRowsBuffered > 1 && m_hstmt != SQL_NULL_HANDLE && m_pRecords == NULL)
        {
        m_nRowsBuffered = nRowsBuffered;
        return (TRUE);
        }

    return (FALSE);
    }

BOOL CListCtlList::SetExecutionMode
    (
    BOOL bBlocking
    )
    {
    if (m_hstmt != SQL_NULL_HANDLE && m_pRecords == NULL)
        {
        m_bBlocking = bBlocking;
        return (TRUE);
        }

    return (FALSE);
    }

BOOL CListCtlList::Query
    (
    BOOL bReconstruct
    )
    {
    UINT        nRow;
    UINT        nCol;

    UINT        nParam;
    LV_ITEM     lvitem;

    if (m_hstmt == SQL_NULL_HANDLE)
        {
        return (FALSE);
        }

    if (m_bBlocking == FALSE)
        {
        SQLCloseCursor(m_hstmt);
        SQLEndTran(SQL_HANDLE_DBC, m_hdbc, SQL_ROLLBACK);
        }

    m_nGridRows = m_nRowsBuffered + 1;

    if (m_bParamsBound == FALSE)
        {
        if (m_listParams.GetCount())
            {
            for (nParam = 0; nParam < (UINT) m_listParams.GetCount(); nParam++)
                {
                m_listParams.GetAt(m_listParams.FindIndex(nParam))->Bind(
                    m_hstmt, nParam + 1);
                }
            }
        
        m_bParamsBound = TRUE;
        }

    if (ExecuteStatement() == FALSE)
        {
        return (FALSE);
        }

    if (m_listColAttr.IsEmpty())
        {
        if (GetColumnAttributes() == FALSE)
            {
            return (FALSE);
            }

        m_pBindOffsets = new UINT[m_nCols];
        m_pIndOffsets  = new UINT[m_nCols];
        if (m_pBindOffsets == NULL || m_pIndOffsets == NULL)
            {
            return (FALSE);
            }

        // GetBindOffsets uses ColAttr column lengths to determine the offsets
        //  for bound values and the total length of a row of bound values
        //  (m_cbRow).
        GetBindOffsets();
        }

    // Check to see if we need to determine the shape of the result set.
    if (bReconstruct == TRUE)
        {
        CColAttr*   pColAttr;
        LV_COLUMN   lvcol;

        TEXTMETRIC  tm;
        CDC*        pdc = m_pList->GetDC();
        pdc->GetTextMetrics(&tm);
        m_pList->ReleaseDC(pdc);

        // Set grid column widths and headings.
        for (nCol = 0; nCol < m_nCols; nCol++)
            {
            pColAttr = m_listColAttr.GetAt(m_listColAttr.FindIndex(nCol));

            lvcol.cx = (int) (tm.tmAveCharWidth * 
                max(pColAttr->GetNameLen(), pColAttr->GetDisplaySize()));
            lvcol.pszText = (PTSTR) (pColAttr->GetName());
		    lvcol.fmt = LVCFMT_LEFT;
            lvcol.mask = LVCF_WIDTH | LVCF_FMT | LVCF_TEXT;

            m_pList->InsertColumn((int) nCol, &lvcol);
            }
        }

    if (m_pRecords == NULL)
        {
        m_pRecords = new BYTE[m_cbRow * m_nRowsBuffered];
        if (m_pRecords == NULL)
            {
            return (FALSE);
            }

        if (BindColumns() == FALSE)
            {
            return (FALSE);
            }
        }

    if (FetchRecords() == FALSE)
        {
        return (FALSE);
        }

    // Remove all rows to start with
    this->Empty();

    m_nRowsInFirstPage = m_nRowsFetched;

    // Set grid values.
    for (nRow = 0; nRow < m_nRowsFetched; nRow++)
        {
        lvitem.mask = LVIF_TEXT;
        lvitem.iItem = (int) nRow;
        for (nCol = 0; nCol < m_nCols; nCol++)
            {
            lvitem.pszText = (PTSTR) (m_pRecords + (m_cbRow * nRow) + m_pBindOffsets[nCol]);
            lvitem.iSubItem = (int) nCol;
            (nCol == 0 ? m_pList->InsertItem(&lvitem) :
                m_pList->SetItem(&lvitem));
            }
        }

    return (TRUE);
    }

void CListCtlList::Empty()
    {
    m_pList->DeleteAllItems();
    }

BOOL CListCtlList::ExecuteStatement()
    {
    // Set up statement handle. Either reset defaults to create a blocking
    //  statement or set statement attributes to create a static,
    //  nonscrollable cursor.
    if (m_bBlocking == TRUE)
        {
        SQLSetStmtAttr(m_hstmt, SQL_ATTR_ROW_ARRAY_SIZE,
            (SQLPOINTER) 1, SQL_IS_INTEGER);
        SQLSetStmtAttr(m_hstmt, SQL_ATTR_CURSOR_TYPE,
            (SQLPOINTER) SQL_CURSOR_FORWARD_ONLY, SQL_IS_INTEGER);
        }
    else
        {
        SQLSetStmtAttr(m_hstmt, SQL_ATTR_ROW_ARRAY_SIZE,
            (SQLPOINTER) (m_nRowsBuffered), SQL_IS_INTEGER);
        SQLSetStmtAttr(m_hstmt, SQL_ATTR_CURSOR_TYPE,
            (SQLPOINTER) SQL_CURSOR_STATIC, SQL_IS_INTEGER);
        }

    // Use direct execution. Developer can optimize by using a SQL Server
    //  stored procedure.
    if (SQL_SUCCEEDED(SQLExecDirect(m_hstmt,
        (SQLTCHAR*) ((LPCTSTR) m_strCommand), SQL_NTS)))
        {
        return (TRUE);
        }

    DisplayErrors(SQL_HANDLE_STMT, m_hstmt);
    return (FALSE);
    }

BOOL CListCtlList::GetColumnAttributes()
    {
    SQLSMALLINT nCols;
    SQLSMALLINT nCol;
    CColAttr*   pColAttr;

    if (SQLNumResultCols(m_hstmt, &nCols) != SQL_SUCCESS)
        {
        DisplayErrors(SQL_HANDLE_STMT, m_hstmt);
        return (FALSE);
        }

    if (nCols == 0)
        {
        return (FALSE);
        }

    for (nCol = 0; nCol < nCols; nCol++)
        {
        pColAttr = new CColAttr(m_hstmt, nCol + 1);
        if (SQL_SUCCEEDED(pColAttr->GetAttr()))
            {
            m_listColAttr.AddTail(pColAttr);
            }
        else
            {
            return (FALSE);
            }
        }

    m_nCols = (UINT) nCols;
    return (TRUE);
    }

void CListCtlList::GetBindOffsets()
    {
    UINT        nCol;

    m_cbRow = 0;

    for (nCol = 0; nCol < m_nCols; nCol++)
        {
        m_pBindOffsets[nCol] = m_cbRow;
        m_cbRow += 
            (m_listColAttr.GetAt(m_listColAttr.FindIndex(nCol)))->GetDisplaySize();
        m_pIndOffsets[nCol] = m_cbRow;
        m_cbRow += sizeof(SQLINTEGER);
        }
    }

// Bind columns, then set statement to bind by row for array fetches.
BOOL CListCtlList::BindColumns()
    {
    UINT        nCol;

    for (nCol = 0; nCol < m_nCols; nCol++)
        {
        if (!SQL_SUCCEEDED(SQLBindCol(m_hstmt, nCol + 1,
            SQL_C_TCHAR, (SQLPOINTER) (m_pRecords + m_pBindOffsets[nCol]),
            (SQLINTEGER) (m_listColAttr.GetAt(
                m_listColAttr.FindIndex(nCol)))->GetDisplaySize(),
            (SQLINTEGER*) (m_pRecords + m_pIndOffsets[nCol]))))
            {
            DisplayErrors(SQL_HANDLE_STMT, m_hstmt);
            return (FALSE);
            }
        }

    SQLSetStmtAttr(m_hstmt, SQL_ATTR_ROW_BIND_TYPE,
        (SQLPOINTER) m_cbRow, SQL_IS_UINTEGER);

    return (TRUE);
    }

BOOL CListCtlList::FetchRecords()
    {
    BOOL    bRet;

    // If the statement is blocking, then cursor type and scrollability are
    //  already set for a default result set. Fetching tickles the row array
    //  size to do fast block fetches, so we need to reset it here to get
    //  back to default set processing.
    if (m_bBlocking == TRUE)
        {
        SQLSetStmtAttr(m_hstmt, SQL_ATTR_ROW_ARRAY_SIZE,
            (SQLPOINTER) m_nRowsBuffered, SQL_IS_INTEGER);

        bRet = SQL_SUCCEEDED(SQLFetch(m_hstmt));
        }
    else
        {
        bRet = SQL_SUCCEEDED(SQLFetchScroll(m_hstmt, SQL_FETCH_FIRST, 0));
        }

    if (bRet == FALSE)
        {
        DisplayErrors(SQL_HANDLE_STMT, m_hstmt);
        }

    // We fetch only the row buffer size of rows from a blocking hstmt. That
    //  may have exhausted the fetch and freed the connection, but it might
    //  not have. Close the cursor to ensure that the connection's free for
    //  other users.
    if (m_bBlocking == TRUE)
        {
        SQLCloseCursor(m_hstmt);
        SQLEndTran(SQL_HANDLE_DBC, m_hdbc, SQL_ROLLBACK);
        }

    return (bRet);
    }

