// CParam.cpp -- implementation of the ODBC parameter class. Parameters can
//  provide data at execution and can receive output from SQL Server stored
//  procedures that return data.
//
// This file is part of Microsoft SQL Server online documentation.
// Copyright (C) 1992-1997 Microsoft Corporation. All rights reserved.
//
// This source code is an intended supplement to the Microsoft SQL
// Server online references and related electronic documentation.
#include "stdafx.h"
#include "Param.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

IMPLEMENT_DYNAMIC(CParam, CObject)

// Construct from a standard C character ptr. Supports same types as 
//  MFC string class constructor.
CParam::CParam(PTSTR pStr, SQLSMALLINT eSQLType, UINT cbSQLCol,
                       UINT nSQLDigits)
    {
    m_cbValue    = _tcslen(pStr);
    m_cbSQLCol   = max(m_cbValue, cbSQLCol);
    m_nSQLDigits = nSQLDigits;
    m_pValue     = (void*) pStr;

    m_eCType     = SQL_C_TCHAR;
    m_eSQLType   = eSQLType;
    m_eParamType = SQL_PARAM_INPUT;

    m_bIsCString = FALSE;
    }

// Construct from a pointer to an unsigned char (BYTE) string. Supports
//  data types like VARBINARY and LONGVARBINARY.
CParam::CParam(PBYTE pByte, UINT cbStr, SQLSMALLINT eSQLType)
    {
    m_eCType     = SQL_C_CHAR;
    m_eSQLType   = eSQLType;

    m_bIsCString = FALSE;

    m_cbValue    = m_cbSQLCol = cbStr;
    m_nSQLDigits = 0;
    m_pValue     = (void*) pByte;
    }

// CParam construct from CString object.
CParam::CParam(CString* pCString, SQLSMALLINT eSQLType, UINT cbSQLCol,
           UINT nSQLDigits)
    {
    m_eCType     = SQL_C_CHAR;
    m_eSQLType   = eSQLType;
    m_eParamType = SQL_PARAM_INPUT;
    
    m_cbValue    = pCString->GetLength();
    m_cbSQLCol   = max(m_cbValue, cbSQLCol);
    m_nSQLDigits = nSQLDigits;
    m_pValue     = (void*) pCString;

    m_bIsCString = TRUE;
    }
        
// Construct from integers (signed or unsigned). Either parameter supports
//  SQL data types of SQL_INTEGER, SQL_SMALLINT, and SQL_TINYINT.
CParam::CParam(int* pInt, SQLSMALLINT eSQLType)
    {
    m_eSQLType   = eSQLType;
    m_eCType     = SQL_C_SLONG;
    m_eParamType = SQL_PARAM_INPUT;

    m_cbSQLCol   = 0;
    m_nSQLDigits = 0;

    m_bIsCString = FALSE;

    m_cbValue    = sizeof(int);
    m_pValue     = (void*) pInt;
    }

CParam::CParam(UINT* pUint, SQLSMALLINT eSQLType)
    {
    m_eSQLType   = eSQLType;
    m_eCType     = SQL_C_ULONG;
    m_eParamType = SQL_PARAM_INPUT;

    m_cbSQLCol   = 0;
    m_nSQLDigits = 0;

    m_bIsCString = FALSE;

    m_cbValue    = sizeof(UINT);
    m_pValue     = (void*) pUint;
    }

// Floating point (approximate numeric) data types.
CParam::CParam(float* pFloat)
    {
    m_eSQLType   = SQL_REAL;
    m_eCType     = SQL_C_FLOAT;
    m_eParamType = SQL_PARAM_INPUT;

    m_bIsCString = FALSE;

    m_cbValue    = sizeof(float);
    m_pValue     = (void*) pFloat;
    }

CParam::CParam(double* pDouble)
    {
    m_eSQLType   = SQL_DOUBLE;
    m_eCType     = SQL_C_DOUBLE;
    m_eParamType = SQL_PARAM_INPUT;

    m_bIsCString = FALSE;

    m_cbValue    = sizeof(double);
    m_pValue     = (void*) pDouble;
    }

// Bind the parameter for use.
BOOL CParam::Bind(SQLHSTMT hStmt, UINT nPos)
    {
    void*       pValue;

    if (m_bIsNull == TRUE)
        {
        m_Indicator = SQL_NULL_DATA;
        }
    else if (m_eSQLType == SQL_CHAR || m_eSQLType == SQL_VARCHAR)
        {
        UINT    cbValue = 0;
        
        m_Indicator = SQL_NTS;
        
        if (m_bIsCString)
            {
            cbValue = ((CString*) m_pValue)->GetLength();
            }
        else 
            {
            cbValue = strlen((PCSTR) m_pValue);
            }

        if (cbValue > m_cbValue)
            {
            m_cbValue = cbValue;
            m_cbSQLCol = max(m_cbValue, m_cbSQLCol);
            }
        }
    else if (m_eSQLType == SQL_LONGVARCHAR || m_eSQLType == SQL_LONGVARBINARY)
        {
        m_Indicator = SQL_LEN_DATA_AT_EXEC((SQLINTEGER) m_cbValue);
        }
    else
        {
        m_Indicator = m_cbValue;
        }

    // If the string is long, send a reference to ourselves.
    //  SQL Server will send us an SQL_NEED_DATA error on execution
    //  and we'll use this reference to provide the data at that 
    //  time.
    if (m_cbValue > MAX_SHORT_STR)
        {
        pValue = ((void*) this);
        }
    else if (m_bIsCString == TRUE)
        {
        pValue = (void*) (((CString*) m_pValue)->GetBuffer(m_cbValue + 1));
        }
    else
        {
        pValue = m_pValue;
        }

    return (SQL_SUCCEEDED(SQLBindParameter(hStmt, (SQLSMALLINT) nPos,
        m_eParamType, m_eCType, m_eSQLType, m_cbSQLCol, m_nSQLDigits, 
        pValue, m_cbValue, &m_Indicator)));
    }

// Return data garnered from an I/O or output-type parameter.
BOOL CParam::GetData(CString* psValue)
    {
    ASSERT (m_eParamType != SQL_PARAM_INPUT);

    PBYTE   pValue;

    if (m_eParamType != SQL_PARAM_INPUT)
        {
        return (FALSE);
        }
try
    {
    pValue = (PBYTE) psValue->GetBufferSetLength((int) m_Indicator);
    }

catch(CMemoryException* e)
    {
    e->Delete();
    return (FALSE);
    }

    switch (m_eSQLType)
        {
        case SQL_CHAR:
        case SQL_VARCHAR:
        case SQL_LONGVARCHAR:
            {
            // strcpy the value to the string since we're null terminated; use
            //  default length setting in the string on release.
            strcpy((PSTR) pValue, (PSTR) m_pValue);
            psValue->ReleaseBuffer();
            }
            break;

        default:
            {
            // memcpy the binary data to the string; then set the length of the
            //  CString through call to release.
            memcpy(pValue, (PBYTE) m_pValue, m_Indicator);
            psValue->ReleaseBuffer((int) m_Indicator);
            }
        }

    return (TRUE);

    }

BOOL CParam::GetData(PBYTE pValue, UINT cbValue, UINT* cbValueOut)
    {
    ASSERT (m_eParamType != SQL_PARAM_INPUT);

    if (m_eParamType != SQL_PARAM_INPUT)
        {
        return (FALSE);
        }

    if (cbValue < (UINT) m_Indicator)
        {
        *cbValueOut = cbValue;
        }
    else
        {
        *cbValueOut = (UINT) m_Indicator;
        }

    memcpy(pValue, (PBYTE) m_pValue, *cbValueOut);

    return (TRUE);
    }

// Set the type; override the default parameter type of input.
void CParam::SetType(int eParamType)
    {
    ASSERT (eParamType == SQL_PARAM_INPUT || eParamType == SQL_PARAM_OUTPUT ||
        eParamType == SQL_PARAM_INPUT_OUTPUT);

    m_eParamType = (SQLSMALLINT) eParamType;
    }

#ifdef _DEBUG
// Everything should be set, including length, just call SQLPutData.
BOOL CParam::PutData(SQLHSTMT hStmt)
    {
    SQLRETURN   fRet;

    fRet = SQLPutData(hStmt, m_pValue, m_cbValue);
    return (fRet == SQL_SUCCESS);
    }

// Nondebug version is inline.
void CParam::SetNull()
    {
    m_bIsNull = TRUE;
    }

// Return the length of data retrieved from an output type parameter.
UINT CParam::GetDataLength()
    {
    ASSERT (m_eParamType != SQL_PARAM_INPUT);

    if (m_eParamType == SQL_PARAM_INPUT)
        {
        return ((UINT) -1);
        }

    return ((UINT) m_Indicator);
    }
#endif

// CParam destruction.
CParam::~CParam()
    {
    }

// Debug help
#ifdef _DEBUG
void CParam::AssertValid() const
{
	CObject::AssertValid();
}

void CParam::Dump(CDumpContext& dc) const
{
	CObject::Dump(dc);
}
#endif //_DEBUG
