//****************************************************************************
//        Copyright (c) 1988-1996 Microsoft Corporation.
//
// @File: ODBCConn.cpp
//
// Purpose:
//    This module contains the CODBCConnection object routines used to manage
//    ODBC connections.
//
// Notes:
//    
//    
// History:
//     
//     @Version: Sphinx
// @EndHeader@
//****************************************************************************
#include <stdio.h>
#include <string.h>                
#include <windows.h>
#include <tchar.h>
#include <odbcconn.h>
#include <odbcss.h>

#ifndef MB_SERVICE_NOTIFICATION
#define MB_SERVICE_NOTIFICATION          0x00200000L
#endif

#define RESOURCE_DLL 	"REPLRES.DLL"

//
// Function: CODBCConnection
//
// Description:
//
//        Constructor for CODBCConnection object.
//
// Returns:
//    
//        Pointer to CODBCConnection object
//
// Notes:
//
//
CODBCConnection::CODBCConnection()
{
    memset(this, 0, sizeof(CODBCConnection));
    m_NumErrors=0;
    m_henv = SQL_NULL_HENV;
    m_hdbc = SQL_NULL_HDBC;
    m_hstmt = SQL_NULL_HSTMT;
    m_Type = SQL_SERVER;          
    m_QueryTimeout = DEFAULT_QUERY_TIMEOUT;
    m_LoginTimeout = DEFAULT_LOGIN_TIMEOUT;
    m_SecurityMode = SECURITY_MODE_STANDARD;           
    m_dwReplOleErrorCode = 0;

	//////////////////////////////////////////////////////////
    //get the host name automatically
    // BUGBUG - may need to remove getcomputername out of the constructor

    TCHAR  szBuffer[ MAX_COMPUTERNAME_LENGTH + 1 ];
    DWORD bufsize = sizeof( szBuffer )/sizeof(TCHAR);
   	if( GetComputerName( szBuffer, &bufsize ) )
   		{
     	DWORD chars_to_copy = min( bufsize, sizeof( m_szHost )/sizeof(TCHAR)-1 );
    	CopyMemory( m_szHost, szBuffer, chars_to_copy *sizeof(TCHAR) );
    	m_szHost[ chars_to_copy ] = _T('\0');
    	}    
    else
    	memset(m_szHost, 0, sizeof(m_szHost));	
    //////////////////////////////////////////////////////////

    m_hr = NOERROR;
    m_SQLServerVersion = SRV_VERSION_NONE;
}

//
// Function: ~CODBCConnection
//
// Description:
//
//        Destructor for CODBCConnection object.
//
// Returns:
//    
//        VOID
//
// Notes:
//
//
CODBCConnection::~CODBCConnection()
{

	if (m_szErrorCommand != 0)
	{
		delete[] m_szErrorCommand;
		m_szErrorCommand = 0;
	}

     Disconnect();
}

//
// Function: Connect
//
// Description:
//
//        This routine connects to a data source using either a direct connect or
//        DSN.
//
// Returns:
//    
//        NOERROR or E_FAIL    or ODBCCONN_E_RETRY    
//
// Notes:
//
//

HRESULT CODBCConnection::Connect(IConnectionInfo *pConnectionInfo)
{
    RETCODE return_code;
    TCHAR  	szConnectString[1024];
    DWORD   value = SQL_IS_ON;
    SWORD 	len;


    // All the parameters needed to make a connection are inherited from the connection info object
    pConnectionInfo->GetODBCEnvironmentHandle(&m_henv);
    pConnectionInfo->GetDataSourceType((REPOLE_DATASOURCE_TYPE *)&m_Type);
    pConnectionInfo->GetSecurityMode((REPOLE_SECURITY_TYPE *)&m_SecurityMode);
    pConnectionInfo->GetPacketSize(&m_PacketSize);
    pConnectionInfo->GetLoginTimeout((LONG *)&m_LoginTimeout);
    pConnectionInfo->GetQueryTimeout((LONG *)&m_QueryTimeout);


    pConnectionInfo->GetConnectName(m_szName, MAX_NAME);
    pConnectionInfo->GetServerName(m_szConnectName, MAX_NAME);
    pConnectionInfo->GetHostName(m_szHost, MAX_NAME);
    pConnectionInfo->GetServerName(m_szServerName, MAX_NAME);
	pConnectionInfo->GetServerRole(m_szServerRole, MAX_NAME);
    pConnectionInfo->GetInternetAddress(m_szAddress, MAX_NAME);
    pConnectionInfo->GetInternetNetwork(m_szNetwork, MAX_NAME);
    pConnectionInfo->GetLogin(m_szLoginName, MAX_NAME);
    pConnectionInfo->GetLogin(m_szUserName, MAX_NAME);
    pConnectionInfo->GetPassword(m_szPassword, MAX_NAME);
    pConnectionInfo->GetDatabase(m_szDatabase, MAX_NAME);
    pConnectionInfo->GetDatasourcePath(m_szDatabasePath, MAX_PATH);
    pConnectionInfo->GetApplicationName(m_szApplication, MAX_NAME);

    /* 
    ** Setting the RemoteServerName same as ServerName allows this connection
    ** to be tagged as a replication login - This ensures that the
    ** NOT FOR REPLICATION triggers do not fire during a MERGE and 
    ** affects other special replication login behavior.
    */
    _tcscpy(m_szRemServerName, m_szServerName);
    //_tcscpy(m_szApplication, (LPCTSTR)pConnectionInfo->m_szApplication);
    

    
    // Open polling servers
    //
    return_code = SQLAllocConnect(m_henv, &m_hdbc); /* Connection handle */
    if (RC_NOTSUCCESSFUL(return_code)){
        goto FAILURE;
    }

    return_code = SQLSetConnectOption(m_hdbc, SQL_LOGIN_TIMEOUT,
        m_LoginTimeout);
    if (RC_NOTSUCCESSFUL(return_code)){
        goto FAILURE;
    }

	return_code = SQLSetConnectOption(m_hdbc, SQL_PACKET_SIZE, m_PacketSize) ;
    if (RC_NOTSUCCESSFUL(return_code)){
        goto FAILURE;
    }

	// Don't convert MBCS data from source to destination code pages!
    return_code = SQLSetConnectOption(m_hdbc, SQL_COPT_SS_TRANSLATE,
        SQL_XL_OFF);
    if (RC_NOTSUCCESSFUL(return_code)){
        goto FAILURE;
    }

 	// If there is a DSN use it to connect, else use the SQL Server driver.
	//
    switch(m_Type)
    {
        case ODBC_DATA_SOURCE:
		    return_code = SQLConnect(m_hdbc, (SQLTCHAR *)m_szConnectName,
			    (SWORD)_tcslen(m_szConnectName), (SQLTCHAR *)m_szLoginName,
    			(SWORD)_tcslen(m_szLoginName), (SQLTCHAR *)m_szPassword,
	    		(SWORD)_tcslen(m_szPassword));
			if (RC_NOTSUCCESSFUL(return_code))
			{
				goto FAILURE;
			}
            break;

        case JET_DATABASE:
           _stprintf (szConnectString, 
			_T("driver={Microsoft Access Driver (*.mdb)};uid=%s;pwd=%s;DBQ=%s"),
			m_szLoginName, m_szPassword, m_szDatabasePath); 

    
		    return_code = SQLDriverConnect(m_hdbc,
			    NULL,
			    (SQLTCHAR *)szConnectString,
			    (SWORD)_tcslen(szConnectString),
			    NULL,
			    0,
			    NULL,
			    SQL_DRIVER_NOPROMPT);
            if (RC_NOTSUCCESSFUL(return_code))
            {
			    goto FAILURE;
		    }
           break;

        default: // SQL Server
    		return_code = SQLSetConnectOption(m_hdbc, SQL_PACKET_SIZE,
	    		m_PacketSize);
    		if (RC_NOTSUCCESSFUL(return_code))
			{
				GetError();
				goto FAILURE;
			}
            if (SECURITY_MODE_INTEGRATED == m_SecurityMode || 
				SECURITY_MODE_NTAUTHENTICATION == m_SecurityMode)
	    	{
		    	// Force integrated security usage
    			//
	    		//return_code = SQLSetConnectOption(m_hdbc,
		    	//	SQL_INTEGRATED_SECURITY, (UDWORD)value);
				return_code = SQLSetConnectAttr(m_hdbc, 
					SQL_INTEGRATED_SECURITY, (SQLPOINTER)value, SQL_IS_INTEGER);
     			if (RC_NOTSUCCESSFUL(return_code))
				{
					GetError();
					goto FAILURE;
				}

				if (SECURITY_MODE_NTAUTHENTICATION == m_SecurityMode)
				{
					/*
					return_code = SQLSetConnectOption(m_hdbc,
		    			SQL_SVR2SVR_RPC, (UDWORD) SQL_SVR2SVR_RPC_ON);
     				if (RC_NOTSUCCESSFUL(return_code))
					{
						GetError();
						goto FAILURE;
					}
					*/
				}



   			}

			// Enable connection for bcp
			return_code = SQLSetConnectAttr(m_hdbc,
			   SQL_COPT_SS_BCP, (SQLPOINTER) SQL_BCP_ON, SQL_IS_INTEGER);
    		if (RC_NOTSUCCESSFUL(return_code))
			{
				GetError();
				goto FAILURE;
			}

            // always set the REPL_SERVER_NAME if we have specified a
            // remote server name.  ( this should only be done by the distributor )
            // setting a remote server name will cause replicated log records generated
            // by this connection to be tagged with the publishing server's name,
            // allowing us to perform loopback detection during distribution.

            if( m_szRemServerName[0] )
		    {
			    return_code |= SQLSetConnectOption(m_hdbc, SQL_REPL_SERVER_NAME,
				    (UDWORD)m_szRemServerName);
			    if (RC_NOTSUCCESSFUL(return_code)){
				    goto FAILURE;
			    }
		    }

            // If an no explict network and address are provided use connection name
            if( !m_szNetwork[0] )
            {
                _stprintf (szConnectString, 
			    _T("driver={SQL Server};uid=%s;pwd=%s;server=%s;wsid=%s;app=%s"),
			    m_szLoginName, m_szPassword, m_szConnectName, m_szHost, m_szApplication);
            }
            else
            {

                _stprintf (szConnectString, 
			    _T("driver={SQL Server};uid=%s;pwd=%s;server=%s;wsid=%s;app=%s;network=%s;address=%s"),
			    m_szLoginName, m_szPassword, m_szConnectName, m_szHost, 
			    m_szApplication, m_szNetwork, m_szAddress);
            }

			return_code = SQLDriverConnect(m_hdbc,
				NULL,
				(SQLTCHAR *)szConnectString,
				(SWORD)_tcslen(szConnectString),
				NULL,
				0,
				NULL,
				SQL_DRIVER_NOPROMPT);

			if (RC_NOTSUCCESSFUL(return_code)){
				goto FAILURE;
			}
    
			// Set database to use if it is specified.
			if(m_szDatabase[0] != _T('\0'))
			{
				m_hr = ChangeDatabase(m_szDatabase);	
				if (NOERROR != m_hr)
				{
					// UNDONE LoadMessage(REPL_BADDB, m_szDatabase, m_szConnectName); 
				    return (m_hr);
				}
			}

            break;
    }
	
	// Create a statement handle for this server
	//
	return_code = SQLAllocStmt(m_hdbc, &m_hstmt);
	if (RC_NOTSUCCESSFUL(return_code))
	{
		goto FAILURE;
	}
	
	return_code = SQLSetStmtOption (m_hstmt, SQL_QUERY_TIMEOUT, m_QueryTimeout);
	if (RC_NOTSUCCESSFUL(return_code))
	{
		if(m_Type == SQL_SERVER)
			goto FAILURE;
	}

	// BUGBUG: Temporary, Set ANSI NULL option
	// This should be removed after 7.0 engine remember ANSI_NULLS
	// settings of stored procedures
	if (m_Type == SQL_SERVER)
	{		
		m_hr = ExecuteDirect(_T("SET ANSI_NULLS OFF"));
		if (NOERROR != m_hr)
			goto FAILURE;
	}

    return_code = SQLGetInfo(m_hdbc, SQL_DBMS_NAME,
        &m_szDbmsName, MAX_TEXT, &len);
    
    if (RC_NOTSUCCESSFUL(return_code)){
        GetError();
        goto FAILURE;
    }

	if (m_Type == SQL_SERVER)
		{
		// it fails for jet, so do it only for sqlserver
	    ExecuteDirect(_T("set quoted_identifier on\n"));
	    }

    return (m_hr);

FAILURE:
	
	// UNDONE LoadMessage(REPL_BADCONN, m_szConnectName);
    GetError();
    return (m_hr);
}


//
// Function: Disconnect
//
// Description:
//
//        This routine disconnects from a data source.
//
// Returns:
//    
//        NOERROR or E_FAIL    or ODBCCONN_E_RETRY    
//
// Notes:
//
//
HRESULT CODBCConnection::Disconnect()
{
    RETCODE        return_code;

    // Display Message
    //

    // In this function, ignore all errors and proceed;
    if (m_hstmt)
    {
        return_code = SQLFreeStmt(m_hstmt, SQL_DROP);
        m_hstmt = SQL_NULL_HSTMT;
    }

    if (m_hdbc)
    {
        // Do a rollback to insure a transactions is not still in progress
        //
        return_code = SQLTransact (m_henv, m_hdbc, SQL_ROLLBACK);

        return_code = SQLDisconnect(m_hdbc);

		return_code = SQLFreeConnect(m_hdbc);

        m_hdbc = SQL_NULL_HDBC;
    }

    return (m_hr);
}



HRESULT CODBCConnection::ChangeDatabase(TCHAR * szDatabase)
{
	RETCODE return_code;
	_tcsncpy(m_szDatabase, szDatabase, sizeof(m_szDatabase)/sizeof(TCHAR) -1);
	return_code = SQLSetConnectOption(m_hdbc, SQL_CURRENT_QUALIFIER,
		(UDWORD)m_szDatabase);
	if (RC_NOTSUCCESSFUL(return_code)){
		GetError();
		goto FAILURE;
	}

FAILURE:
    return (m_hr);
}

//
// Function: ClearErrors
//
// Description:
//
//        This routine resets the m_NumErrors and m_hr fields.
//
// Returns:
//    
//    	nothing   
//
//
VOID CODBCConnection::ClearErrors()
{
	m_NumErrors = 0;
	m_hr = NOERROR;
}

//
// Function: GetError
//
// Description:
//
//        This routine connects load errors into the m_Error array.
//
// Returns:
//    
//        NOERROR or E_FAIL    or ODBCCONN_E_RETRY    
//
// Notes:
//
//        The maximum number of errors maintained is defined by MAX_ERRORS.
//
HRESULT CODBCConnection::GetError()
{
    int		i;
	BOOL	bError;
	BOOL	bNetLibError = FALSE;	// Net library errors will always be returned with an information 
									// error.  This is to signal when a NetLib error is detected.
							
 	TCHAR *	chSource;
 	TCHAR *	chSourceEnd;
 	TCHAR *	chMessage ;
	DWORD	dwMessage;
	BYTE	len;

	for (i = m_NumErrors; i < MAX_ERRORS; i++) 
    {
		bError = FALSE;

		// Get statement level errors
		bError = SQLError(m_henv, m_hdbc, m_hstmt, (SQLTCHAR*)m_Error[i].szSqlState, 
			&m_Error[i].NativeError, (SQLTCHAR *)m_Error[i].szErrMsg, 
			SQL_MAX_MESSAGE_LENGTH, &m_Error[i].wErrMsgLen);
		

		// No statement level errors, get connection level error
		// Note: bcp errors will come here.
		if (SQL_SUCCESS != bError)
		{
			bError = SQLError(m_henv, m_hdbc, SQL_NULL_HSTMT, (SQLTCHAR*)m_Error[i].szSqlState, 
				&m_Error[i].NativeError, (SQLTCHAR *)m_Error[i].szErrMsg, 
				SQL_MAX_MESSAGE_LENGTH, &m_Error[i].wErrMsgLen);
		}

		if (SQL_SUCCESS != bError)
			break;

		// Determine Source of Error
		chSource = _tcsrchr(m_Error[i].szErrMsg, _T('['));
		if (chSource)
		{
			chSource++;
			chSourceEnd = _tcschr(chSource, _T(']'));
			len = chSourceEnd-chSource;

			// Return Driver Manager errors as ODBC source errors
			//BUG use resource string
			if (!_tcsncmp (chSource, _T("ODBC Driver Manager"), _tcslen(_T("ODBC Driver Manager"))))
			{
				m_Error[i].SourceType = REPL_SOURCE_ODBC;
			}
			// Special SQL Server Drive handling for Net-Library errors				
			// The first error message is returned as a Driver error and the follow up 
			// information error is returned as a Net-Library error
			else if (!_tcsncmp (chSource, _T("DBNMPNTW"), _tcslen(_T("DBNMPNTW"))) ||
					 !_tcsncmp (chSource, _T("DBMSSOCN"), _tcslen(_T("DBMSSOCN"))) ||
					 !_tcsncmp (chSource, _T("DBMSSPXN"), _tcslen(_T("DBMSSPXN"))) ||
					 !_tcsncmp (chSource, _T("DBMSRPCN"), _tcslen(_T("DBMSRPCN"))))
			{
				if (bNetLibError)
				{
					m_Error[i].SourceType = REPL_SOURCE_NETLIB;
				}
				else
				{
					m_Error[i].SourceType = REPL_SOURCE_ODBC;
					// Return first error as SQL Server Driver 
					_tcscpy(m_Error[i].szSourceName, _T("ODBC SQL Server Driver"));
					bNetLibError = TRUE;
				}
			}
			// All other sources without _T("Driver") in them are assumed to be Data Source errors
			else if (!_tcsstr(chSource, _T("Driver")))
			{
				m_Error[i].SourceType = REPL_SOURCE_SQLSERVER; 
				_tcscpy(m_Error[i].szSourceName, m_szConnectName);
			}
			else // All other errors are ODBC
			{
				m_Error[i].SourceType = REPL_SOURCE_ODBC;
			}

			// Load Source name, if not already loaded
			if (!*m_Error[i].szSourceName)
			{
				memcpy(m_Error[i].szSourceName, chSource, len*sizeof(TCHAR));
			}

			// Remove ODBC Header [..][..] and parse out Net-Library information 
			chMessage = _tcsrchr(m_Error[i].szErrMsg, _T(']'));
			dwMessage = _tcslen(chMessage);
			if (chMessage)
			{
				memcpy (m_Error[i].szErrMsg, ++chMessage, dwMessage*sizeof(TCHAR));
				m_Error[i].szErrMsg[dwMessage] = _T('\0');
			}
		}

		// Remove terminating newline characters
		for( WORD ThisByte = m_Error[i].wErrMsgLen-1; ThisByte ; ThisByte -- )
		{
			TCHAR* pThisChar =  m_Error[i].szErrMsg + ThisByte;
			if( _T('\r') == *pThisChar || _T('\n') == *pThisChar )
			{
				*pThisChar = _T('\0');
			}
			else
			{
				break;  // break when non newline character encountered in backwards scan
			}
		}

        // Construct Native Error string
        //
        _stprintf (m_Error[i].szNativeError, _T("%d"), m_Error[i].NativeError);

		// For skip and go, in DistCmds, if we find 20524 or 20525 or 20558 or 20526 or 20527, we log to history 
		// and to the new MSrepl_alerts table, and go on. -RamS.
		if (!_tcsicmp(m_szDbmsName, _T("Microsoft SQL Server")) && 
			(m_Error[i].NativeError == 20524 || m_Error[i].NativeError == 20525 || 
			m_Error[i].NativeError == 20558 || m_Error[i].NativeError == 20526 || 
			m_Error[i].NativeError == 20527 || m_Error[i].NativeError == 20578))
		{
			m_NumErrors++;

			// 20524 and 20525 are validation failures, whereas 20526 and 20527 are validation passing codes.
			if  (0 == m_SkipFailureLevel && 
				(m_Error[i].NativeError == 20524 || 
				m_Error[i].NativeError == 20525))
			{
				m_hr = E_FAIL;
			}

			//if (m_Error[i].NativeError == 20578)
			//{
				//m_bShutdownAgentAfterCommit = TRUE;
			//}
		
			continue;
		}

		// Ignore all distribution subsystem raiserror messages
        //
        if (m_Error[i].NativeError == 14150 || m_Error[i].NativeError == 14151 ||
            m_Error[i].NativeError == 14152 || m_Error[i].NativeError == 14153 ||
			m_Error[i].NativeError == 20574 || m_Error[i].NativeError == 20575 ||
			m_Error[i].NativeError == 14158 )
		{
			m_NumErrors++;
			// Will happen only while adding history which may call sp_MSrepl_raiserror. Not while executing the repl commands/batches.
			break;
		}


        // Ignore all general warning messages
        //
        if (!_tcsncmp (m_Error[i].szSqlState, _T("01000"), _tcslen(_T("01000"))))
        {
            /* Reuse the m_Error slot */
            i--;
            continue;
        }

		// Increment after ignoring warnings
        m_NumErrors++;
        
        // If server failure or operation aborted, retry
        //

        if( IsRetryableError( &m_Error[i], &m_dwReplOleErrorCode ) )
        {                            
            m_hr = ODBCCONN_E_RETRY;
			// Keep the retry return status and we don't care following errors.
			break;
        }
		else if (_tcsncmp (m_Error[i].szSqlState, _T("01000"), _tcslen(_T("01000")))) // ignore information errors
        {
            m_dwReplOleErrorCode = ROLE_ODBC_ERROR;
            m_hr = E_FAIL;
        }
    }

    return(m_hr);
}


//
// Function: ConstructOutcome
//
// Description:
//
//        This routine constructs an outcome string for the CODBCConnection.
//
// Returns:
//    
//        NOERROR or E_FAIL    or ODBCCONN_E_RETRY    
//
// Notes:
//
//
HRESULT CODBCConnection::ConstructOutcome()
{

	// Only return 1 message in Outcome message, Error logging will keep track of other
	// messages
    if (!m_cbLen && m_hr != NOERROR && 
		m_hr != ODBCCONN_E_CANCEL)
	{
		// Load Object level error message
		Concat(m_szMsg);
		
		// Concat First ODBC Error message ( only if there are any )
		//
		if (!m_cbLen && m_NumErrors > 0)
			Concat(m_Error[0].szErrMsg);
	}
	return m_hr;
}
//
// Function: CODBCConnection::GetErrorCodeText
//
// Description:
//
//        This routine passes back the error info in the buffers provided by the caller.
//		(Since the m_Error should really be private...)
//
// Returns:
//    
//
// Notes:
//
//
VOID CODBCConnection::GetErrorCodeText(SDWORD *pNativeErrorCode, TCHAR *szNativeErrorText, int BuffLen)
{
	*pNativeErrorCode = m_Error[0].NativeError;
	_tcsncpy(szNativeErrorText, m_Error[0].szErrMsg, BuffLen);
}


//
// Function: ExecuteDirect
//
// Description:
//
//        This routine is a wrapper around ODBC's ExecuteDirect.
//
// Returns:
//    
//        NOERROR or E_FAIL    or ODBCCONN_E_RETRY    
//
// Notes:
//
//
HRESULT CODBCConnection::ExecuteDirect(TCHAR * szBuffer)
{
    RETCODE        return_code;

    // Set data flag
    m_bData = FALSE;

	// Clear the error command array buffer
	if (m_szErrorCommand != 0)
	{
		delete[] m_szErrorCommand;
		m_szErrorCommand = 0;
	}

	if (NULL == szBuffer || _T('\0') == *szBuffer)
		goto Exit;

	return_code = SQLExecDirect(m_hstmt, (SQLTCHAR *)szBuffer, SQL_NTS);
   
	if ( (SQL_SERVER == m_Type && SQL_SUCCESS != return_code) || RC_NOTSUCCESSFUL(return_code) )
	{
		// For SQL Server, we want to do this even for SQL_SUCCESS_WITH_INFO, e.g. when sp_table_validation fails and raises sev 10 error
		GetError();
		
		if(m_hr != NOERROR && m_hr != ODBCCONN_E_CANCEL)
		{
			m_szErrorCommand = new TCHAR[_tcslen(szBuffer) + 1 ];
			_tcscpy(m_szErrorCommand, szBuffer);
		}
	}

Exit:
	
	return(m_hr);
}

// close the cursor...
LONG 
CODBCConnection::CloseCursor()
{
    RETCODE        return_code;

	SQLFreeStmt(m_hstmt, SQL_RESET_PARAMS);
    return_code = SQLFreeStmt(m_hstmt, SQL_CLOSE);
        if (RC_NOTSUCCESSFUL(return_code))
    {
        GetError();
    }

    return m_hr;
}
    
//
// Function: Prepare
//
// Description:
//
//        This routine is a wrapper around ODBC's Prepare.
//
// Returns:
//    
//        NOERROR or E_FAIL    or ODBCCONN_E_RETRY    
//
// Notes:
//
//
HRESULT CODBCConnection::Prepare(TCHAR * szBuffer)
{
    RETCODE        return_code;


	// Clear the error command array buffer
	if (m_szErrorCommand != 0)
	{
		delete[] m_szErrorCommand;
		m_szErrorCommand = 0;
	}

	if (NULL == szBuffer || _T('\0') == *szBuffer)
		goto Exit;

    return_code = SQLPrepare(m_hstmt, (SQLTCHAR *)szBuffer, SQL_NTS);
    if (RC_NOTSUCCESSFUL(return_code))
    {
        GetError();

    	// Record the last cmd that failed.
		m_szErrorCommand = new TCHAR[_tcslen(szBuffer) + 1 ];
		_tcscpy(m_szErrorCommand, szBuffer);
	}

Exit:

    return(m_hr);
}

//
// Function: Execute
//
// Description:
//
//        This routine is a wrapper around ODBC's Execute.
//
// Returns:
//    
//        NOERROR or E_FAIL    or ODBCCONN_E_RETRY    
//
// Notes:
//
//
HRESULT CODBCConnection::Execute()
{
    RETCODE        return_code;

    // Set data flag
    //
    m_bData = FALSE;

	// Ignore the execute if the task has been shutdown.
	//

    return_code = SQLExecute(m_hstmt);
    if (RC_NOTSUCCESSFUL(return_code))
    {
        GetError();
    }

    return(m_hr);
}

//
// Function: Fetch
//
// Description:
//
//        This routine is a wrapper around ODBC's Fetch.
//
// Returns:
//    
//        NOERROR or E_FAIL    or ODBCCONN_E_RETRY    
//
// Notes:
//
//
LONG 
CODBCConnection::Fetch()
    {
    RETCODE        return_code;

    return_code = SQLFetch(m_hstmt);
    if (RC_NOTSUCCESSFUL(return_code))
        {
        if (return_code == SQL_NO_DATA_FOUND)
            {
            m_bData = FALSE;
            return(m_hr);
            }

        GetError();
        
        goto Error;
        }

    m_bData = TRUE;

Error:
    return(m_hr);
    }


//
// Function: MoreResults
//
// Description:
//
//        This routine is a wrapper around ODBC's SQLMoreResults.
//
// Returns:
//    
//        NOERROR or E_FAIL    or ODBCCONN_E_RETRY    
//
// Notes:
//
//
HRESULT CODBCConnection::MoreResults()
{
    RETCODE        return_code;

    while(TRUE)
    {
           // Check that the results in the batch are all successful. If not
        // the whole batch is 'rolled back'.
        //
        return_code = SQLMoreResults(m_hstmt);

        // Check if there are more results
        //
        if (RC_NOTSUCCESSFUL(return_code))
            break;
    }
       
    if (return_code != SQL_NO_DATA_FOUND)
    {
        GetError();
        if (!m_fNoErrorAssert)
        	SQLCancel(m_hstmt);
    }
    
    return(m_hr);
}

//
// Function: GetNextResultSet
//
// Description:
//
//        This routine is a wrapper around ODBC's SQLMoreResults.
//
// Returns:
//    
//        NOERROR or E_FAIL    or ODBCCONN_E_RETRY    
//
// Notes:
//
//
HRESULT CODBCConnection::GetNextResultSet()
{
    RETCODE        return_code;

    // Check that the results in the batch are all successful. If not
    // the whole batch is 'rolled back'.
    //
    return_code = SQLMoreResults(m_hstmt);

    // Check if there are more results
    //
    if (RC_NOTSUCCESSFUL(return_code))
	{
		if (return_code != SQL_NO_DATA_FOUND)
		{
			GetError();
			if (!m_fNoErrorAssert)
        		SQLCancel(m_hstmt);
		}
	}
    
    return(m_hr);
}

//
// Function: GetColumnData
//
// Description:
//
//        Gets data for a column from the previously executed sql-command...
//
// Returns:
//    
//
// Notes:
//
//
LONG 
CODBCConnection::GetColumnData(int iCol, int ColType, TCHAR *pData, SDWORD cbMax, SDWORD *pcbActual)
    {
    RETCODE        retCode;
    
    retCode = SQLGetData(m_hstmt, iCol, ColType, pData, cbMax, pcbActual);

    if (RC_NOTSUCCESSFUL(retCode))
        {
        GetError();
        }
        
    return(m_hr);
    }


//
// Function: GetColumnCount
//
// Description:
//
//        Gets datatype for a column from the previously executed sql-command...
//
// Returns:
//    pColType receives the datatype
//	  retCode is a status value
//
// Notes:
//
//
LONG 
CODBCConnection::GetColumnCount(LONG * plCount)
    {
    RETCODE	rc;
	rc = SQLColAttributes(m_hstmt,
					0,
					SQL_COLUMN_COUNT,
					NULL,
					NULL,
					NULL,
					plCount);
    if (RC_NOTSUCCESSFUL(rc))
        {
        GetError();
        }
        
    return(m_hr);
    }

//
// Function: GetDatatype
//
// Description:
//
//        Gets datatype for a column from the previously executed sql-command...
//
// Returns:
//    pColType receives the datatype
//	  retCode is a status value
//
// Notes:
//
//
LONG 
CODBCConnection::GetColumnDatatype(int iCol, LONG * plDataType)
    {
    RETCODE	rc;
	rc = SQLColAttributes(m_hstmt,
					iCol,
					SQL_COLUMN_TYPE,
					NULL,
					NULL,
					NULL,
					plDataType);
    if (RC_NOTSUCCESSFUL(rc))
        {
        GetError();
        }
        
    return(m_hr);
    }
LONG 
CODBCConnection::GetColumnUpdatable(int iCol, LONG * plColumnUpdatable, LONG *plColumnIdentity)
    {
    RETCODE	rc;
	rc = SQLColAttributes(m_hstmt,
					iCol,
					SQL_COLUMN_UPDATABLE,
					NULL,
					NULL,
					NULL,
					plColumnUpdatable);
    if (RC_NOTSUCCESSFUL(rc))
        {
        GetError();
        }

   	rc = SQLColAttributes(m_hstmt,
					iCol,
					SQL_COLUMN_AUTO_INCREMENT,
					NULL,
					NULL,
					NULL,
					plColumnIdentity);
    if (RC_NOTSUCCESSFUL(rc))
        {
        GetError();
        }

    return(m_hr);
    }

LONG 
CODBCConnection::GetColumnName(int iCol, LPTSTR szColumnName, int cbBufferLength)
    {
	short	len;
    RETCODE	rc;
	rc = SQLColAttributes(m_hstmt,
					iCol,
					SQL_COLUMN_NAME,
					szColumnName,
					cbBufferLength,
					&len,
					NULL);
    if (RC_NOTSUCCESSFUL(rc))
        {
        GetError();
        }
        
    return(m_hr);
    }
LONG 
CODBCConnection::GetColumnLength(int iCol, LONG * plColumnLength)
    {
    RETCODE	rc;
  
	rc = SQLColAttributes(m_hstmt,
					iCol,
					SQL_COLUMN_LENGTH,
					NULL,
					NULL,
					NULL,
					plColumnLength);
    if (RC_NOTSUCCESSFUL(rc))
        {
        GetError();
        }
        
    return(m_hr);
    }
LONG 
CODBCConnection::GetColumnScale(int iCol, LONG * plColumnScale)
    {
    RETCODE	rc;
  
	rc = SQLColAttributes(m_hstmt,
					iCol,
					SQL_COLUMN_SCALE,
					NULL,
					NULL,
					NULL,
					plColumnScale);
    if (RC_NOTSUCCESSFUL(rc))
        {
        GetError();
        }
        
    return(m_hr);
    }

LONG 
CODBCConnection::GetColumnPrecision(int iCol, LONG * plColumnPrecision)
    {
    RETCODE	rc;
  
	rc = SQLColAttributes(m_hstmt,
					iCol,
					SQL_COLUMN_PRECISION,
					NULL,
					NULL,
					NULL,
					plColumnPrecision);
    if (RC_NOTSUCCESSFUL(rc))
        {
        GetError();
        }
        
    return(m_hr);
    }

LONG 
CODBCConnection::GetDisplayLength(int iCol, LONG * plDisplayLength)
    {
    RETCODE	rc;
  
	rc = SQLColAttributes(m_hstmt,
					iCol,
					SQL_COLUMN_DISPLAY_SIZE,
					NULL,
					NULL,
					NULL,
					plDisplayLength);
    if (RC_NOTSUCCESSFUL(rc))
        {
        GetError();
        }
        
    return(m_hr);
    }


HRESULT CODBCConnection::BeginTran()
{
	RETCODE return_code;
	
	return_code = SQLSetConnectOption(m_hdbc, SQL_AUTOCOMMIT, SQL_AUTOCOMMIT_OFF);
	if (RC_NOTSUCCESSFUL(return_code))
	{
		GetError();
	}	
	return m_hr;
}


HRESULT CODBCConnection::Commit()
{
    RETCODE return_code;
    
    // Commit job
    //
    return_code = SQLTransact(SQL_NULL_HENV, m_hdbc, SQL_COMMIT);
    if (RC_NOTSUCCESSFUL(return_code)){
        GetError();
        goto FAILURE;
	}

FAILURE:
    return m_hr;
}


HRESULT CODBCConnection::Rollback()
{
    RETCODE return_code;
    
    // Commit job
    //
    return_code = SQLTransact(m_henv, m_hdbc, SQL_ROLLBACK);
    if (RC_NOTSUCCESSFUL(return_code))
    {
        GetError();
    }

    return m_hr;
}

HRESULT CODBCConnection::BindParameter(int iCol, int ParamType, int CType, int SqlType, 
    UDWORD cbColDef, SWORD ibScale, LPVOID pvData, SDWORD cbMax, SDWORD *pcbActual)
    {
    RETCODE    retCode;

	if (SQL_C_TCHAR == CType && (SQL_CHAR == SqlType || SQL_VARCHAR == SqlType) && 0 == cbColDef && 
		!_tcsicmp(m_szDbmsName, _T("Oracle")))
	{
		cbColDef = _tcslen((TCHAR *) pvData);
	}
    
    retCode = SQLBindParameter(m_hstmt, iCol, ParamType, CType, 
             SqlType, cbColDef, ibScale, pvData, cbMax, pcbActual);

    if (RC_NOTSUCCESSFUL(retCode))
        {
        GetError();
        }

    return(m_hr);
    }

HRESULT CODBCConnection::GetData(int iCol, int CType, LPVOID pvData, SDWORD cbMax, SDWORD *pcbActual)
    {
    RETCODE    retCode;
    
    retCode = SQLGetData(m_hstmt, iCol, CType, pvData, cbMax, pcbActual);
    if (RC_NOTSUCCESSFUL(retCode))
        {
        GetError();
        }

    return(m_hr);
    }
    
/////////////////////////////////////////////////////////////////////////////
//  Function:   BOOL CODBCConnection::IsRetryableError( ODBC_ERROR* pError, DWORD* pOleErrCode )
//  Created:    11/19/96 EricRe
//  Purpose:    decide whether or not an error is retryable and return the
//              REPLOLE error code
//  Returns:    TRUE if error is retryable, FALSE if not
/////////////////////////////////////////////////////////////////////////////

BOOL CODBCConnection::IsRetryableError( ODBC_ERROR* pError, DWORD* pOleErrCode )
{
    BOOL fRetval = TRUE;

    if( !pError || !pOleErrCode )
    {
        fRetval = FALSE; // fall through, return FALSE
    }
    else if (!_tcsncmp( pError->szSqlState, _T("08"), _tcslen(_T("08"))) )
    {
        *pOleErrCode = ROLE_CONNECTION_FAILURE;
    }
    else if( !_tcsncmp( pError->szSqlState, _T("70"), _tcslen(_T("70"))) )
    {
        *pOleErrCode = ROLE_OPERATION_ABORTED;
    }
    else if( !_tcsncmp( pError->szSqlState, _T("S1T00"), _tcslen(_T("S1T00"))) )
    {
        *pOleErrCode = ROLE_CONNECTION_TIMEOUT;
    }
    else if( SQL_SERVER == m_Type || !_tcsicmp( m_szDbmsName, _T("Microsoft SQL Server")) ) 
	{
		if (1205 == pError->NativeError )
		{
			*pOleErrCode = ROLE_DEADLOCK;
		}
		else
		{
            fRetval = FALSE;
		}
    }
    else if( !_tcsicmp( m_szDbmsName, _T("Access")) )
    {
        if( -1304 == pError->NativeError )
        {
            *pOleErrCode = ROLE_TABLE_LOCKED;
        }
        else if( -1302 == pError->NativeError )
        {
            *pOleErrCode = ROLE_CANNOT_LOCK_TABLE;
        }
        else
        {
            fRetval = FALSE;
        }
    }
    else
    {
        fRetval = FALSE;
    }

    return fRetval;
}




//
// Function: LoadMessage
//
// Description:
//      Load a message string and formats it using the specified arguments
//      into m_szMsg.  If there is not enough space the string will be truncated.
//
//
VOID CODBCConnection::LoadMessage(DWORD wID, ...)
{
    HINSTANCE 	hdl;
    TCHAR 		szFormat[MAX_FORMAT_LEN];
    int 		cbNumBytesWritten = 0;        // num bytes written to buffer by _vsnprintf
    va_list 	arg_ptr;                 // pointer to argument list 

    // Get handle of the resource library
    hdl = LoadLibraryA(RESOURCE_DLL);
    if (!hdl)
    {
		UINT uType = MB_SYSTEMMODAL | MB_ICONSTOP | MB_OK;

		OSVERSIONINFO VerInfo;
		VerInfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);

		if( GetVersionEx( &VerInfo ) && 
			VER_PLATFORM_WIN32_NT == VerInfo.dwPlatformId )
		{
			uType |= MB_SERVICE_NOTIFICATION;
		}

		MessageBoxA( NULL, "Invalid handle attempting to load message from resource DLL.", "String Load Error", uType );

        return;
    }
        
    // Load message string
    //
    LoadString(hdl, wID, szFormat, MAX_FORMAT_LEN);

    va_start(arg_ptr, wID);                            
    cbNumBytesWritten = _vsntprintf(m_szMsg, MAX_MESSAGE_LEN - 1, szFormat, arg_ptr);     
    va_end(arg_ptr);

    // Ensure that string is properly null terminated
    //
    if( cbNumBytesWritten > 0 )
    {
        m_szMsg[ cbNumBytesWritten ] = _T('\0');    
    }
    else if( -1 == cbNumBytesWritten )
    {
        m_szMsg[MAX_MESSAGE_LEN] = _T('\0');
    }
    else
    {
        // UNDONE LoadString(hdl, REPL_BAD_MSG, m_szMsg, MAX_MESSAGE_LEN);
    }    

    FreeLibrary( hdl );
}

BOOL CODBCConnection::Concat(TCHAR * szMessage)
{
	if (!_tcslen(szMessage))
		return (TRUE);

	if (_tcslen(szMessage)-1 <= MAX_MESSAGE_LEN - m_cbLen)
	{
		_tcscat (m_szOutcome, szMessage);
		m_cbLen += _tcslen(szMessage);
		return (TRUE);
	}
	return (FALSE);
}


/////////////////////////////////////////////////////////////////////////////
// GetCTypeFromSqlType - Map a SQL type to the C equivalent.
//
// Returns:  The C type
/////////////////////////////////////////////////////////////////////////////
SWORD GetCTypeFromSqlType(SWORD fSqlType)					// SQL type to map to a C type
	{
	switch(fSqlType) {
		case SQL_CHAR:
		case SQL_VARCHAR:
		case SQL_LONGVARCHAR:
		case SQL_DECIMAL:
		case SQL_NUMERIC:
		case SQL_BIGINT:
			return SQL_C_CHAR;
		case SQL_WCHAR:
		case SQL_WVARCHAR:
		case SQL_WLONGVARCHAR:
			return SQL_C_WCHAR;
		case SQL_GUID:
			return SQL_C_GUID;

		case SQL_BIT:
			return SQL_C_BIT;
			
		case SQL_TINYINT:
			return SQL_C_TINYINT;
			
		case SQL_SMALLINT:
			return SQL_C_SHORT;
			
		case SQL_INTEGER:
			return SQL_C_LONG;
		
		case SQL_REAL:
			return SQL_C_FLOAT;
			
		case SQL_FLOAT:
		case SQL_DOUBLE:
			return SQL_C_DOUBLE;
			
		case SQL_BINARY:
		case SQL_VARBINARY:
		case SQL_LONGVARBINARY:
			return SQL_C_BINARY;
			
		case SQL_DATE:
			return SQL_C_DATE;
			
		case SQL_TIME:
			return SQL_C_TIME;
			
		case SQL_TIMESTAMP:
			return SQL_C_TIMESTAMP;
			
		default:
			return SQL_C_DEFAULT;
		}
	}

